import trafaret as t

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2, parse_time


class AllowedTimeUnitsSAFER(object):
    """ Enum for SAFER allowed time units
    """

    MILLISECOND = "MILLISECOND"
    SECOND = "SECOND"
    MINUTE = "MINUTE"
    HOUR = "HOUR"
    DAY = "DAY"
    WEEK = "WEEK"
    MONTH = "MONTH"
    QUARTER = "QUARTER"
    YEAR = "YEAR"

    ALL = t.Enum(MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR)


class RelationshipsConfiguration(APIObject):
    """ A Relationships configuration specifies a set of secondary datasets as well as
    the relationships among them. It is used to configure Feature Discovery for a project
    to generate features automatically from these datasets.

    Attributes
    ----------
    id : str
        the id of the created relationships configuration
    dataset_definitions: list
        each element is a dataset_definitions for a dataset.
    relationships: list
        each element is a relationship between two datasets

    The `dataset_defintions` structure is

    identifier: str
        alias of the dataset (used directly as part of the generated feature names)
    catalog_id: str, or None
        identifier of the catalog item
    catalog_version_id: str
        identifier of the catalog item version
    primary_temporal_key: str, or None
        name of the column indicating time of record creation
    feature_list_id: str, or None
        identifier of the feature list. This decides which columns in the dataset are
        used for feature generation
    snapshot_policy: str
        policy to use  when creating a project or making predictions.
        Must be one of the following values:
        'specified': Use specific snapshot specified by catalogVersionId
        'latest': Use latest snapshot from the same catalog item
        'dynamic': Get data from the source (only applicable for JDBC datasets)
    feature_lists: list
        list of feature list info
    data_source: dict
        data source info if the dataset is from data source
    is_deleted: bool or None
        whether the dataset is deleted or not

    The `data source info` structured is

    data_store_id: str
        the id of the data store.
    data_store_name : str
        the user-friendly name of the data store.
    url : str
        the url used to connect to the data store.
    dbtable : str
        the name of table from the data store.
    schema: str
        schema definition of the table from the data store

    The `feature list info` structure is

    id : str
        the id of the featurelist
    name : str
        the name of the featurelist
    features : list of str
        the names of all the Features in the featurelist
    dataset_id : str
        the project the featurelist belongs to
    creation_date : datetime.datetime
        when the featurelist was created
    user_created : bool
        whether the featurelist was created by a user or by DataRobot automation
    created_by: str
        the name of user who created it
    description : str
        the description of the featurelist.  Can be updated by the user
        and may be supplied by default for DataRobot-created featurelists.
    dataset_id: str
        dataset which is associated with the feature list
    dataset_version_id: str or None
        version of the dataset which is associated with feature list.
        Only relevant for Informative features

    The `relationships` schema is

    dataset1_identifier: str or None
        identifier of the first dataset in this relationship.
        This is specified in the indentifier field of dataset_definition structure.
        If None, then the relationship is with the primary dataset.
    dataset2_identifier: str
        identifier of the second dataset in this relationship.
        This is specified in the identifier field of dataset_definition schema.
    dataset1_keys: list of str (max length: 10 min length: 1)
        column(s) from the first dataset which are used to join to the second dataset
    dataset2_keys: list of str (max length: 10 min length: 1)
        column(s) from the second dataset that are used to join to the first dataset
    time_unit: str, or None
        time unit of the feature derivation window. Supported
        values are MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR.
        If present, the feature engineering Graph will perform time-aware joins.
    feature_derivation_window_start: int, or None
        how many time_units of each dataset's primary temporal key into the past relative
        to the datetimePartitionColumn the feature derivation window should begin.
        Will be a negative integer,
        If present, the feature engineering Graph will perform time-aware joins.
    feature_derivation_window_end: int, or None
        how many timeUnits of each dataset's record
        primary temporal key into the past relative to the datetimePartitionColumn the
        feature derivation window should end.  Will be a non-positive integer, if present.
        If present, the feature engineering Graph will perform time-aware joins.
    feature_derivation_window_time_unit: int or None
        time unit of the feature derivation window. Supported values are
        MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR
        If present, time-aware joins will be used.
        Only applicable when dataset1Identifier is not provided.
    prediction_point_rounding: int, or None
        closest value of prediction_point_rounding_time_unit to round the prediction point
        into the past when applying the feature derivation window. Will be a positive integer,
        if present.Only applicable when dataset1_identifier is not provided.
    prediction_point_rounding_time_unit: str, or None
        time unit of the prediction point rounding. Supported values are
        MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR
        Only applicable when dataset1_identifier is not provided.
    """

    _path = "relationshipsConfigurations/"
    data_source_trafaret = t.Dict(
        {
            t.Key("data_store_name"): t.String,
            t.Key("data_store_id"): t.String,
            t.Key("url"): t.String,
            t.Key("dbtable"): t.String | t.Null,
            t.Key("schema", optional=True): t.String | t.Null,
        }
    )
    feature_list_info = t.Dict(
        {
            t.Key('id'): t.String,
            t.Key('features'): t.List(t.String),
            t.Key('name'): t.String,
            t.Key('description'): t.String,
            t.Key('dataset_id'): t.String,
            t.Key('user_created'): t.Bool,
            t.Key('creation_date'): parse_time,
            t.Key('created_by'): t.String,
            t.Key('dataset_version_id', optional=True): t.String
        }
    )
    dataset_definitions_trafaret = t.Dict(
        {
            t.Key("identifier"): t.String(min_length=3, max_length=20),
            t.Key("catalog_version_id"): t.String,
            t.Key("catalog_id"): t.String,
            t.Key("primary_temporal_key", optional=True): t.String | t.Null,
            t.Key("feature_list_id", optional=True): t.String | t.Null,
            t.Key("snapshot_policy", optional=True, default="latest"): t.Enum(
                "latest", "specified", "dynamic"
            ),
            t.Key('feature_lists', optional=True): t.List(feature_list_info),
            t.Key("data_source", optional=True): data_source_trafaret | t.Null,
            t.Key("is_deleted", optional=True): t.Bool | t.Null,
        }
    )

    relationships_trafaret = t.Dict(
        {
            t.Key("dataset1_identifier", optional=True): t.String | t.Null,
            t.Key("dataset2_identifier"): t.String,
            t.Key("dataset1_keys"): t.List(t.String, min_length=1, max_length=10),
            t.Key("dataset2_keys"): t.List(t.String, min_length=1, max_length=10),
            t.Key("feature_derivation_window_start", optional=True): t.Int(lt=0),
            t.Key("feature_derivation_window_end", optional=True): t.Int(lte=0),
            t.Key(
                "feature_derivation_window_time_unit", optional=True
            ): AllowedTimeUnitsSAFER.ALL,
            t.Key("prediction_point_rounding", optional=True): t.Int(gt=0, lte=30),
            t.Key(
                "prediction_point_rounding_time_unit", optional=True
            ): AllowedTimeUnitsSAFER.ALL,
        }
    )
    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("dataset_definitions"): t.List(
                dataset_definitions_trafaret, min_length=1
            ),
            t.Key("relationships"): t.List(relationships_trafaret, min_length=1),
        }
    ).ignore_extra("*")

    def __init__(self, id, dataset_definitions=None, relationships=None):
        self.id = id
        self.dataset_definitions = dataset_definitions
        self.relationships = relationships

    def __repr__(self):
        return encode_utf8_if_py2(u"{}()".format(self.__class__.__name__))

    @classmethod
    def create(cls, dataset_definitions, relationships):
        """ Create a Relationships Configuration

        Parameters
        ----------
        dataset_definitions: list of dict
            each element is a DatasetDefinition .
            The `DatasetDefinition` schema is

            identifier: str
                alias of the table (used directly as part of the generated feature names)
            catalog_id: str, or None
                identifier of the catalog item
            catalog_version_id: str
                identifier of the catalog item version
            feature_list_id: str, or None
                identifier of the feature list. This decides which columns in the table are
                used for feature generation
            snapshot_policy: str
                policy to use  when creating a project or making predictions.
                Must be one of the following values:
                'specified': Use specific snapshot specified by catalogVersionId
                'latest': Use latest snapshot from the same catalog item
                'dynamic': Get data from the source (only applicable for JDBC datasets)
        relationships: list of dict
            each element is a Relationship between two datasets
            The `Relationship` schema is

            dataset1_identifier: str or None
                identifier of the first dataset in this relationship.
                This is specified in the indentifier field of dataset_definition structure.
                If None, then the relationship is with the primary dataset.
            dataset2_identifier: str
                identifier of the second dataset in this relationship.
                This is specified in the identifier field of dataset_definition schema.
            dataset1_keys: list of str (max length: 10 min length: 1)
                column(s) from the first dataset which are used to join to the second dataset
            dataset2_keys: list of str (max length: 10 min length: 1)
                column(s) from the second dataset that are used to join to the first dataset
            time_unit: str, or None
                time unit of the feature derivation window. Supported
                values are MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR.
                If present, the feature engineering Graph will perform time-aware joins.
            feature_derivation_window_start: int, or None
                how many time_units of each dataset's primary temporal key into the past relative
                to the datetimePartitionColumn the feature derivation window should begin.
                Will be a negative integer,
                If present, the feature engineering Graph will perform time-aware joins.
            feature_derivation_window_end: int, or None
                how many timeUnits of each dataset's record
                primary temporal key into the past relative to the datetimePartitionColumn the
                feature derivation window should end.  Will be a non-positive integer, if present.
                If present, the feature engineering Graph will perform time-aware joins.
            feature_derivation_window_time_unit: int or None
                time unit of the feature derivation window. Supported values are
                MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR
                If present, time-aware joins will be used.
                Only applicable when dataset1Identifier is not provided.
            prediction_point_rounding: int, or None
                closest value of prediction_point_rounding_time_unit to round the prediction point
                into the past when applying the feature derivation window. Will be a positive
                integer, if present.Only applicable when dataset1_identifier is not provided.
            prediction_point_rounding_time_unit: str, or None
                time unit of the prediction point rounding. Supported values are
                MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR
                Only applicable when dataset1_identifier is not provided.

        Returns
        -------
        relationships_configuration: RelationshipsConfiguration
            the created relationships configuration
        """
        payload_data = {
            "dataset_definitions": dataset_definitions,
            "relationships": relationships,
        }
        response = cls._client.post(cls._path, data=payload_data).json()
        return RelationshipsConfiguration.from_server_data(response)

    def get(self):
        """ Retrieve the Relationships configuration for a given id

        Returns
        -------
        relationships_configuration: RelationshipsConfiguration
            The requested relationships configuration

        Raises
        ------
        ClientError
            Raised if an invalid relationships config id is provided.

        Examples
        --------
        .. code-block:: python

            relationships_config = dr.RelationshipsConfiguration(valid_config_id)
            result = relationships_config.get()
            >>> result.id
            '5c88a37770fc42a2fcc62759'
        """
        return self.from_location("{}{}/".format(self._path, self.id))

    def replace(self, dataset_definitions, relationships):
        """ Update the Relationships Configuration which is not used in
        the feature discovery Project

        Parameters
        ----------
        dataset_definitions: list of dict
            each element is a DatasetDefinition .
            The `DatasetDefinition` schema is

            identifier: str
                alias of the table (used directly as part of the generated feature names)
            catalog_id: str, or None
                identifier of the catalog item
            catalog_version_id: str
                identifier of the catalog item version
            feature_list_id: str, or None
                identifier of the feature list. This decides which columns in the table are
                used for feature generation
            snapshot_policy: str
                policy to use  when creating a project or making predictions.
                Must be one of the following values:
                'specified': Use specific snapshot specified by catalogVersionId
                'latest': Use latest snapshot from the same catalog item
                'dynamic': Get data from the source (only applicable for JDBC datasets)
        relationships: list of dict
            each element is a Relationship between two datasets
            The `Relationship` schema is

            dataset1_identifier: str or None
                identifier of the first dataset in this relationship.
                This is specified in the indentifier field of dataset_definition structure.
                If None, then the relationship is with the primary dataset.
            dataset2_identifier: str
                identifier of the second dataset in this relationship.
                This is specified in the identifier field of dataset_definition schema.
            dataset1_keys: list of str (max length: 10 min length: 1)
                column(s) from the first dataset which are used to join to the second dataset
            dataset2_keys: list of str (max length: 10 min length: 1)
                column(s) from the second dataset that are used to join to the first dataset
            time_unit: str, or None
                time unit of the feature derivation window. Supported
                values are MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR.
                If present, the feature engineering Graph will perform time-aware joins.
            feature_derivation_window_start: int, or None
                how many time_units of each dataset's primary temporal key into the past relative
                to the datetimePartitionColumn the feature derivation window should begin.
                Will be a negative integer,
                If present, the feature engineering Graph will perform time-aware joins.
            feature_derivation_window_end: int, or None
                how many timeUnits of each dataset's record
                primary temporal key into the past relative to the datetimePartitionColumn the
                feature derivation window should end.  Will be a non-positive integer, if present.
                If present, the feature engineering Graph will perform time-aware joins.
            feature_derivation_window_time_unit: int or None
                time unit of the feature derivation window. Supported values are
                MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR
                If present, time-aware joins will be used.
                Only applicable when dataset1Identifier is not provided.
            prediction_point_rounding: int, or None
                closest value of prediction_point_rounding_time_unit to round the prediction point
                into the past when applying the feature derivation window. Will be a positive
                integer, if present.Only applicable when dataset1_identifier is not provided.
            prediction_point_rounding_time_unit: str, or None
                time unit of the prediction point rounding. Supported values are
                MILLISECOND, SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, YEAR
                Only applicable when dataset1_identifier is not provided.

        Returns
        -------
        relationships_configuration: RelationshipsConfiguration
            the updated relationships configuration
        """
        payload_data = {
            "datasetDefinitions": dataset_definitions,
            "relationships": relationships,
        }
        url = '{}{}/'.format(self._path, self.id)
        response = self._client.put(url, json=payload_data).json()
        return RelationshipsConfiguration.from_server_data(response)

    def delete(self):
        """
        Delete the Relationships configuration

        Raises
        ------
        ClientError
            Raised if an invalid relationships config id is provided.

        Examples
        --------
        .. code-block:: python

            # Deleting with a valid id
            relationships_config = dr.RelationshipsConfiguration(valid_config_id)
            status_code = relationships_config.delete()
            status_code
            >>> 204
            relationships_config.get()
            >>> ClientError: Relationships Configuration not found
        """
        self._client.delete('{}{}/'.format(self._path, self.id))



from collections import OrderedDict

import mock
import pytest
import responses

from datarobot.enums import VARIABLE_TYPE_TRANSFORM
from tests.utils import request_body_to_json


@pytest.fixture
def batch_type_transform_create_url(unittest_endpoint, project_id):
    return '{}/projects/{}/batchTypeTransformFeatures/'.format(
        unittest_endpoint,
        project_id
    )


@pytest.fixture
def batch_type_transform_create_response(batch_type_transform_create_url, async_url):
    responses.add(
        responses.POST,
        batch_type_transform_create_url,
        status=202,
        body='',
        adding_headers={'Location': async_url}
    )
    responses.add(
        responses.GET,
        async_url,
        status=303,
        body='',
        adding_headers={'Location': '/features/'}
    )


@pytest.mark.parametrize("exception, patch", [
    (TypeError, {"parent_names": []}),
    (ValueError, {"variable_type": "wrong"}),
])
def test_batch_var_type_transform_arguments_validation(project, exception, patch):
    args = OrderedDict({
        'parent_names': ['f1'],
        'variable_type': VARIABLE_TYPE_TRANSFORM.TEXT,
        'prefix': 'prefix',
        'suffix': 'suffix',
    })

    args.update(patch)

    with pytest.raises(exception):
        project.batch_features_type_transform(
            parent_names=args['parent_names'], variable_type=args['variable_type'],
            prefix=args['prefix'], suffix=args['suffix'])


@responses.activate
@pytest.mark.usefixtures('batch_type_transform_create_response')
def test_batch_var_type_transform(project):
    args = [
        ['f1', 'f2'],
        VARIABLE_TYPE_TRANSFORM.NUMERIC,
        'prefix',
    ]

    with mock.patch.object(project, 'get_features') as get_features_mock:
        features = project.batch_features_type_transform(*args)
        assert get_features_mock.called_once
        assert get_features_mock.return_value is features

    payload = request_body_to_json(responses.calls[0].request)

    assert payload['parentNames'] == ['f1', 'f2']
    assert payload['variableType'] == VARIABLE_TYPE_TRANSFORM.NUMERIC
    assert payload['prefix'] == 'prefix'

    assert 'suffix' not in payload


@responses.activate
@pytest.mark.usefixtures('batch_type_transform_create_response')
@pytest.mark.usefixtures('known_warning')
def test_batch_var_type_transform_categorical(project):
    args = [
        ['f1', 'f2'],
        VARIABLE_TYPE_TRANSFORM.CATEGORICAL,
        'prefix',
    ]

    with mock.patch.object(project, 'get_features') as get_features_mock:
        features = project.batch_features_type_transform(*args)
        assert get_features_mock.called_once
        assert get_features_mock.return_value is features

    payload = request_body_to_json(responses.calls[0].request)

    assert payload['parentNames'] == ['f1', 'f2']
    assert payload['variableType'] == VARIABLE_TYPE_TRANSFORM.CATEGORICAL
    assert payload['prefix'] == 'prefix'

    assert 'suffix' not in payload



import os
import unittest
import six.moves.urllib.parse as urllibparse

from datarobot.helpers.eligibility_result import EligibilityResult
from datarobot.helpers.partitioning_methods import (BasePartitioningMethod, GroupTVH,
                                                    DatetimePartitioningSpecification)

_here = os.path.dirname(os.path.abspath(__file__))
FIXTURES_DIR = os.path.join(_here, 'fixtures')


def fixture_file_path(filename):
    return os.path.join(FIXTURES_DIR, filename)


class URLParamsTestCase(unittest.TestCase):

    def assert_has_params(self, url, **params):
        query_string = urllibparse.unquote(url).split('?')[1]
        query_parts = query_string.split('&')
        query_pairs = [param.split('=') for param in query_parts]
        query_params = {pair[0]: pair[1]
                        for pair in query_pairs}

        for key, value in query_params.items():
            self.assertIn(key, query_params,
                          'Expected parameter {} not found'.format(key))
            self.assertEqual(query_params[key], params[key])


class HelpersTestCase(unittest.TestCase):

    def test_cant_create_without_cv(self):
        with self.assertRaises(TypeError):
            BasePartitioningMethod(validation_type='CV')

        with self.assertRaises(TypeError):
            BasePartitioningMethod(cv_method='user')

    def test_partitioning_repr(self):
        p_method = GroupTVH(holdout_pct=1, validation_pct=2,
                            partition_key_cols=['one_feature', 'sec feature'],
                            seed=42)
        assert repr(p_method).startswith('GroupTVH({')

    def test_create_with_right(self):
        p_method = BasePartitioningMethod(validation_type='CV',
                                          cv_method='user')
        self.assertEqual(p_method.validation_type, 'CV')
        self.assertEqual(p_method.cv_method, 'user')

    def test_collect_payload(self):
        p_method = BasePartitioningMethod(validation_type='CV',
                                          cv_method='user')
        p = p_method.collect_payload()
        self.assertTrue('validation_type' in p)
        self.assertTrue('cv_method' in p)
        self.assertTrue('partition_key_cols' in p)
        self.assertEqual(p['validation_type'], 'CV')
        self.assertIsNone(p['partition_key_cols'])

    def test_datetime_collect_payload_unsupervised_mode(self):
        # if unsupervised_mode is not set or False we don't want it in payload
        spec = DatetimePartitioningSpecification('date', unsupervised_mode=False)
        payload = spec.collect_payload()
        assert 'unsupervised_mode' not in payload

        # otherwise it should be present
        spec = DatetimePartitioningSpecification('date', unsupervised_mode=True)
        payload = spec.collect_payload()
        assert payload['unsupervised_mode'] is True


def test_eligiblity_result_repr():
    er = EligibilityResult(False)
    representation = str(er)
    assert 'supported=False' in representation
    assert 'reason=\'\'' in representation
    assert 'context=\'\'' in representation

    er = EligibilityResult(True, reason='reason', context='context')
    representation = str(er)
    assert 'supported=True' in representation
    assert 'reason=\'reason\'' in representation
    assert 'context=\'context\'' in representation


def test_eligibility_result_with_japanese(unicode_string):
    er = EligibilityResult(True, reason=unicode_string, context=unicode_string)
    print(er)  # test __repr__



import os
import contextlib2
import trafaret as t
from requests_toolbelt import MultipartEncoder

from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2
from datarobot.utils.pagination import unpaginate


class CustomModelFileItem(APIObject):
    """A file item attached to a DataRobot custom model version.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        id of the file item
    file_name: str
        name of the file item
    file_path: str
        path of the file item
    file_source: str
        source of the file item
    created_at: str, optional
        ISO-8601 formatted timestamp of when the version was created
    """

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("file_name"): t.String(),
            t.Key("file_path"): t.String(),
            t.Key("file_source"): t.String(),
            t.Key("created", optional=True) >> "created_at": t.String(),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        id,
        file_name,
        file_path,
        file_source,
        created_at=None,
    ):
        self.id = id
        self.file_name = file_name
        self.file_path = file_path
        self.file_source = file_source
        self.created_at = created_at


class CustomModelVersion(APIObject):
    """A version of a DataRobot custom model.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        id of the custom model version
    custom_model_id: str
        id of the custom model
    version_minor: int
        a minor version number of custom model version
    version_major: int
        a major version number of custom model version
    is_frozen: bool
        a flag if the custom model version is frozen
    items: List[CustomModelFileItem]
        a list of file items attached to the custom model version
    label: str, optional
        short human readable string to label the version
    description: str, optional
        custom model version description
    created_at: str, optional
        ISO-8601 formatted timestamp of when the version was created
    """

    _path = "customModels/{}/versions/"

    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("custom_model_id"): t.String(),
            t.Key("version_minor"): t.Int(),
            t.Key("version_major"): t.Int(),
            t.Key("is_frozen"): t.Bool(),
            t.Key("items"): t.List(CustomModelFileItem.schema),
            t.Key("label", optional=True): t.String(
                max_length=50, allow_blank=True
            ) | t.Null(),
            t.Key("description", optional=True): t.String(
                max_length=10000, allow_blank=True
            ) | t.Null(),
            t.Key("created", optional=True) >> "created_at": t.String(),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}({!r})".format(self.__class__.__name__, self.label or self.id)
        )

    def _set_values(
        self,
        id,
        custom_model_id,
        version_minor,
        version_major,
        is_frozen,
        items,
        label=None,
        description=None,
        created_at=None,
    ):
        self.id = id
        self.custom_model_id = custom_model_id
        self.version_minor = version_minor
        self.version_major = version_major
        self.is_frozen = is_frozen
        self.items = [CustomModelFileItem(**item) for item in items]
        self.label = label
        self.description = description
        self.created_at = created_at

    @classmethod
    def create_clean(
        cls,
        custom_model_id,
        is_major_update=True,
        folder_path=None,
        files=None,
    ):
        """Create a custom model version without files from previous versions.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        is_major_update: bool
            the flag defining if a custom model version
            will be a minor or a major version.
            Default to `True`
        folder_path: str, optional
            the path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative
            to a folder path
        files: list, optional
            the list of tuples, where values in each tuple are the local filesystem path and
            the path the file should be placed in the model.
            Example:
            [("/home/user/Documents/myModel/file1.txt", "file1.txt"),
            ("/home/user/Documents/myModel/folder/file2.txt", "folder/file2.txt")]

        Returns
        -------
        CustomModelVersion
            created custom model version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        return cls._create('post', custom_model_id, is_major_update, folder_path, files)

    @classmethod
    def create_from_previous(
        cls,
        custom_model_id,
        is_major_update=True,
        folder_path=None,
        files=None,
        files_to_delete=None,
    ):
        """Create a custom model version containing files from a previous version.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        is_major_update: bool, optional
            the flag defining if a custom model version
            will be a minor or a major version.
            Default to `True`
        folder_path: str, optional
            the path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative
            to a folder path
        files: list, optional
            the list of tuples, where values in each tuple are the local filesystem path and
            the path the file should be placed in the model.
            Example:
            [("/home/user/Documents/myModel/file1.txt", "file1.txt"),
            ("/home/user/Documents/myModel/folder/file2.txt", "folder/file2.txt")]
        files_to_delete: list, optional
            the list of a file items ids to be deleted
            Example: ["5ea95f7a4024030aba48e4f9", "5ea6b5da402403181895cc51"]

        Returns
        -------
        CustomModelVersion
            created custom model version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        if files_to_delete:
            upload_data = [('filesToDelete', file_id) for file_id in files_to_delete]
        else:
            upload_data = None
        return cls._create(
            'patch', custom_model_id, is_major_update, folder_path, files, upload_data
        )

    @classmethod
    def _create(
        cls,
        method,
        custom_model_id,
        is_major_update,
        folder_path=None,
        files=None,
        extra_upload_data=None
    ):
        url = cls._path.format(custom_model_id)

        with contextlib2.ExitStack() as stack:
            upload_data = [('isMajorUpdate', str(is_major_update))]

            if folder_path:
                for root_path, _, file_paths in os.walk(folder_path):
                    for path in file_paths:
                        file_path = os.path.join(root_path, path)
                        file = stack.enter_context(open(file_path, "rb"))

                        upload_data.append(('file', (os.path.basename(file_path), file)))
                        upload_data.append(('filePath', os.path.relpath(file_path, folder_path)))

            if files:
                for file_path, upload_file_path in files:
                    file = stack.enter_context(open(file_path, "rb"))

                    upload_data.append(('file', (os.path.basename(upload_file_path), file)))
                    upload_data.append(('filePath', upload_file_path))

            if extra_upload_data:
                upload_data += extra_upload_data

            encoder = MultipartEncoder(fields=upload_data)
            headers = {'Content-Type': encoder.content_type}
            response = cls._client.request(method, url, data=encoder, headers=headers)
            return cls.from_server_data(response.json())

    @classmethod
    def list(cls, custom_model_id):
        """List custom model versions.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model

        Returns
        -------
        List[CustomModelVersion]
            a list of custom model versions

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = cls._path.format(custom_model_id)
        data = unpaginate(url, None, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, custom_model_id, custom_model_version_id):
        """Get custom model version by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        custom_model_version_id: str
            the id of the custom model version to retrieve

        Returns
        -------
        CustomModelVersion
            retrieved custom model version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        url = cls._path.format(custom_model_id)
        path = "{}{}/".format(url, custom_model_version_id)
        return cls.from_location(path)

    def download(self, file_path):
        """Download custom model version.

        .. versionadded:: v2.21

        Parameters
        ----------
        file_path: str
            path to create a file with custom model version content

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        url = self._path.format(self.custom_model_id)
        path = "{}{}/download/".format(url, self.id)

        response = self._client.get(path)
        with open(file_path, "wb") as f:
            f.write(response.content)

    def update(self, description):
        """Update custom model version properties.

        .. versionadded:: v2.21

        Parameters
        ----------
        description: str
            new custom model version description

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = {"description": description}

        url = self._path.format(self.custom_model_id)
        path = "{}{}/".format(url, self.id)

        response = self._client.patch(path, data=payload)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    def refresh(self):
        """Update custom model version with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = self._path.format(self.custom_model_id)
        path = "{}{}/".format(url, self.id)

        response = self._client.get(path)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))



# flake8: noqa

from ._version import __version__

from .enums import (
    SCORING_TYPE,
    QUEUE_STATUS,
    AUTOPILOT_MODE,
    VERBOSITY_LEVEL,
    TARGET_TYPE,
)
from .client import Client
from .errors import AppPlatformError
from .helpers import *
from .models import (
    Project,
    Model,
    PrimeModel,
    BlenderModel,
    FrozenModel,
    DatetimeModel,
    RatingTableModel,
    Ruleset,
    ModelJob,
    Blueprint, BlueprintTaskDocument, BlueprintChart, ModelBlueprintChart,
    DatasetFeaturelist,
    Featurelist,
    ModelingFeaturelist,
    Feature,
    ModelingFeature,
    DatasetFeature,
    ExternalScores,
    ExternalLiftChart,
    ExternalMulticlassLiftChart,
    ExternalRocCurve,
    ExternalConfusionChart,
    FeatureHistogram,
    DatasetFeatureHistogram,
    PredictJob,
    Job,
    PredictionDataset,
    ImportedModel,
    PrimeFile,
    ReasonCodesInitialization,
    ReasonCodes,
    Predictions,
    PredictionExplanationsInitialization,
    PredictionExplanations,
    RatingTable,
    ShapImpact,
    ShapMatrix,
    ShapMatrixJob,
    SharingAccess,
    TrainingPredictions,
    TrainingPredictionsJob,
    ModelRecommendation,
    DataDriver,
    Dataset,
    DatasetDetails,
    DataStore,
    DataSource,
    DataSourceParameters,
    ComplianceDocumentation,
    ComplianceDocTemplate,
    CalendarFile,
    PredictionServer,
    Deployment,
    FeatureEngineeringGraph,
    SecondaryDatasetConfigurations,
    BatchPredictionJob,
    Credential,
    InteractionFeature,
    ExecutionEnvironment,
    ExecutionEnvironmentVersion,
    CustomInferenceImage,
    CustomModelVersion,
    CustomModelTest,
    CustomInferenceModel,
    ExternalResidualsChart,
    FeatureImpactJob,
    PayoffMatrix,
    RelationshipsConfiguration,
)



import trafaret as t

from datarobot.models.api_object import APIObject
from ..utils import deprecation_warning, encode_utf8_if_py2


class BlueprintTaskDocument(APIObject):
    """ Document describing a task from a blueprint.

    Attributes
    ----------
    title : str
        Title of document.
    task : str
        Name of the task described in document.
    description : str
        Task description.
    parameters : list of dict(name, type, description)
        Parameters that task can receive in human-readable format.
    links : list of dict(name, url)
        External links used in document
    references : list of dict(name, url)
        References used in document. When no link available url equals None.
    """
    _converter = t.Dict({
        t.Key('title'): t.String,
        t.Key('task'): t.String(allow_blank=True),
        t.Key('description'): t.String(allow_blank=True),
        t.Key('parameters'): t.List(t.Dict({
            t.Key('name'): t.String,
            t.Key('type'): t.String,
            t.Key('description'): t.String,
        }).ignore_extra('*')),
        t.Key('links'): t.List(t.Dict({
            t.Key('name'): t.String,
            t.Key('url'): t.String,
        }).ignore_extra('*')),
        t.Key('references'): t.List(t.Dict({
            t.Key('name'): t.String,
            # from_api method drops None, so we need this when there is no url
            t.Key('url', optional=True, default=None): t.String | t.Null,
        }).ignore_extra('*')),
    }).ignore_extra('*')

    def __init__(self, title=None, task=None,
                 description=None, parameters=None,
                 links=None, references=None):
        self.title = title
        self.task = task
        self.description = description
        self.parameters = parameters
        self.links = links
        self.references = references

    def __repr__(self):
        return encode_utf8_if_py2(u'BlueprintTaskDocument({})'.format(self.title))


class BlueprintChart(APIObject):
    """ A Blueprint chart that can be used to understand data flow in blueprint.

    Attributes
    ----------
    nodes : list of dict (id, label)
        Chart nodes, id unique in chart.
    edges : list of tuple (id1, id2)
        Directions of data flow between blueprint chart nodes.
    """

    _converter = t.Dict({
        t.Key('nodes', optional=True): t.List(t.Dict({
            t.Key('id'): t.String,
            t.Key('label'): t.String
        })),
        t.Key('edges', optional=True): t.List(t.Tuple(t.String, t.String)),
    })

    def __init__(self, nodes, edges):
        self.nodes = nodes
        self.edges = edges

    def __repr__(self):
        return encode_utf8_if_py2(u'BlueprintChart({} nodes, {} edges)'.format(len(self.nodes),
                                                                               len(self.edges)))

    @classmethod
    def get(cls, project_id, blueprint_id):
        """ Retrieve a blueprint chart.

        Parameters
        ----------
        project_id : str
            The project's id.
        blueprint_id : str
            Id of blueprint to retrieve chart.

        Returns
        -------
        BlueprintChart
            The queried blueprint chart.
        """
        url = 'projects/{}/blueprints/{}/blueprintChart/'.format(project_id, blueprint_id)
        return cls.from_location(url)

    def to_graphviz(self):
        """ Get blueprint chart in graphviz DOT format.

        Returns
        -------
        unicode
            String representation of chart in graphviz DOT language.
        """
        digraph = u'digraph "Blueprint Chart" {'
        digraph += u'\ngraph [rankdir=LR]'
        for node in self.nodes:
            digraph += u'\n{id} [label="{label}"]'.format(id=node['id'], label=node['label'])
        for edge in self.edges:
            digraph += u'\n{id0} -> {id1}'.format(id0=edge[0], id1=edge[1])
        digraph += u'\n}'
        return digraph


class ModelBlueprintChart(BlueprintChart):
    """ A Blueprint chart that can be used to understand data flow in model.
    Model blueprint chart represents reduced repository blueprint chart with
    only elements that used to build this particular model.

    Attributes
    ----------
    nodes : list of dict (id, label)
        Chart nodes, id unique in chart.
    edges : list of tuple (id1, id2)
        Directions of data flow between blueprint chart nodes.
    """

    def __repr__(self):
        return encode_utf8_if_py2(
            u'ModelBlueprintChart({} nodes, {} edges)'.format(len(self.nodes),
                                                              len(self.edges))
        )

    @classmethod
    def get(cls, project_id, model_id):
        """ Retrieve a model blueprint chart.

        Parameters
        ----------
        project_id : str
            The project's id.
        model_id : str
            Id of model to retrieve model blueprint chart.

        Returns
        -------
        ModelBlueprintChart
            The queried model blueprint chart.
        """
        url = 'projects/{}/models/{}/blueprintChart/'.format(project_id,
                                                             model_id)
        return cls.from_location(url)


class Blueprint(APIObject):
    """ A Blueprint which can be used to fit models

    Attributes
    ----------
    id : str
        the id of the blueprint
    processes : list of str
        the processes used by the blueprint
    model_type : str
        the model produced by the blueprint
    project_id : str
        the project the blueprint belongs to
    blueprint_category : str
        (New in version v2.6) Describes the category of the blueprint and the kind of model it
        produces.
    recommended_featurelist_id: str or null
        (New in v2.18) The ID of the feature list recommended for this blueprint.
        If this field is not present, then there is no recommended feature list.
    """
    _converter = t.Dict({
        t.Key('id', optional=True): t.String(),
        t.Key('processes', optional=True): t.List(t.String()),
        t.Key('model_type', optional=True): t.String(),
        t.Key('project_id', optional=True): t.String(),
        t.Key('blueprint_category', optional=True): t.String(),
        t.Key('monotonic_increasing_featurelist_id', optional=True): t.String(),
        t.Key('monotonic_decreasing_featurelist_id', optional=True): t.String(),
        t.Key('supports_monotonic_constraints', optional=True): t.Bool(),
        t.Key('recommended_featurelist_id', optional=True): t.String(),
    }).allow_extra('*')

    def __init__(self, id=None, processes=None, model_type=None, project_id=None,
                 blueprint_category=None, monotonic_increasing_featurelist_id=None,
                 monotonic_decreasing_featurelist_id=None,
                 supports_monotonic_constraints=None, recommended_featurelist_id=None):
        if isinstance(id, dict):
            deprecation_warning('Blueprint instantiation from a dict',
                                deprecated_since_version='v2.3',
                                will_remove_version='v3.0',
                                message='Use Blueprint.from_data instead')
            self.__init__(**id)
        else:
            self.id = id
            self.processes = processes
            self.model_type = model_type
            self.project_id = project_id
            self.blueprint_category = blueprint_category
            self.monotonic_increasing_featurelist_id = monotonic_increasing_featurelist_id
            self.monotonic_decreasing_featurelist_id = monotonic_decreasing_featurelist_id
            self.supports_monotonic_constraints = supports_monotonic_constraints
            self.recommended_featurelist_id = recommended_featurelist_id

    def __repr__(self):
        return encode_utf8_if_py2(u'Blueprint({})'.format(self.model_type))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    @classmethod
    def get(cls, project_id, blueprint_id):
        """ Retrieve a blueprint.

        Parameters
        ----------
        project_id : str
            The project's id.
        blueprint_id : str
            Id of blueprint to retrieve.

        Returns
        -------
        blueprint : Blueprint
            The queried blueprint.
        """
        url = 'projects/{}/blueprints/{}/'.format(project_id, blueprint_id)
        return cls.from_location(url)

    def get_chart(self):
        """ Retrieve a chart.

        Returns
        -------
        BlueprintChart
            The current blueprint chart.
        """
        return BlueprintChart.get(self.project_id, self.id)

    def get_documents(self):
        """ Get documentation for tasks used in the blueprint.

        Returns
        -------
        list of BlueprintTaskDocument
            All documents available for blueprint.
        """
        url = 'projects/{}/blueprints/{}/blueprintDocs/'.format(self.project_id, self.id)
        return [BlueprintTaskDocument.from_server_data(data) for data in self._server_data(url)]



import time

import trafaret as t

from datarobot.enums import DEFAULT_MAX_WAIT
from datarobot.errors import AsyncProcessUnsuccessfulError
from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


class CustomModelTest(APIObject):
    """An custom model test.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        test id
    dataset_id: str
        id of a dataset used for testing
    dataset_version_id: str
        id of a dataset version used for testing
    custom_model_image_id: str
        id of a custom model image
    overall_status: str
        a string representing testing status.
        Status can be
        - 'not_tested': the check not run
        - 'failed': the check failed
        - 'succeeded': the check succeeded
        - 'warning': the check resulted in a warning, or in non-critical failure
        - 'in_progress': the check is in progress
    detailed_status: dict
        detailed testing status - maps the testing types to their status and message.
        The keys of the dict are one of 'errorCheck', 'nullValueImputation',
        'longRunningService', 'sideEffects'.
        The values are dict with 'message' and 'status' keys.
    created_by: str
        a user who created a test
    completed_at: str, optional
        ISO-8601 formatted timestamp of when the test has completed
    created_at: str, optional
        ISO-8601 formatted timestamp of when the version was created
    """

    _path = "customModelTests/"
    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("dataset_id"): t.String(),
            t.Key("dataset_version_id"): t.String(),
            t.Key("custom_model_image_id"): t.String(),
            t.Key("overall_status"): t.String(),
            t.Key("testing_status") >> "detailed_status": t.Dict(
                {
                    t.Key(test_type): t.Dict(
                        {t.Key("status"): t.String(), t.Key("message"): t.String(allow_blank=True)}
                    ).allow_extra("*")
                    for test_type in [
                        "error_check",
                        "null_value_imputation",
                        "long_running_service",
                        "side_effects",
                    ]
                }
            ).allow_extra("*"),
            t.Key("created_by"): t.String(),
            t.Key("completed_at", optional=True): t.String(allow_blank=True),
            t.Key("created", optional=True) >> "created_at": t.String(),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}({!r})".format(self.__class__.__name__, self.id)
        )

    def _set_values(
        self,
        id,
        dataset_id,
        dataset_version_id,
        custom_model_image_id,
        overall_status,
        detailed_status,
        created_by,
        completed_at=None,
        created_at=None,
    ):
        self.id = id
        self.dataset_id = dataset_id
        self.dataset_version_id = dataset_version_id
        self.custom_model_image_id = custom_model_image_id
        self.overall_status = overall_status
        self.detailed_status = detailed_status
        self.created_by = created_by
        self.completed_at = completed_at
        self.created_at = created_at

    @classmethod
    def create(
            cls,
            custom_model_id,
            custom_model_version_id,
            dataset_id,
            environment_id,
            environment_version_id=None,
            max_wait=DEFAULT_MAX_WAIT,
    ):
        """Create and start a custom model test.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model
        custom_model_version_id: str
            the id of the custom model version
        dataset_id: str
            the id of the testing dataset
        environment_id: str
            the id of the execution environment
        environment_version_id: str, optional
            the id of the execution environment version
        max_wait: int, optional
            max time to wait for a test completion.
            If set to None - method will return without waiting.

        Returns
        -------
        CustomModelTest
            created custom model test

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = {
            "custom_model_id": custom_model_id,
            "custom_model_version_id": custom_model_version_id,
            "dataset_id": dataset_id,
            "environment_id": environment_id,
            "environment_version_id": environment_version_id,
        }
        response = cls._client.post(cls._path, data=payload)

        # at this point custom model test is already created
        custom_model_test = cls.list(custom_model_id)[0]

        if max_wait is None:
            # return without waiting for the test to finish
            return custom_model_test
        else:
            try:
                # wait for the test to finish
                custom_model_test_loc = wait_for_async_resolution(
                    cls._client, response.headers['Location'], max_wait
                )
                return cls.from_location(custom_model_test_loc)
            except AsyncProcessUnsuccessfulError:
                # if the job was aborted server sends appropriate status and
                # `wait_for_async_resolution` raises exception,
                # but the test has been already created, and contains error log,
                # so return the test

                # the test needs some time to update its state
                max_state_wait = 10
                custom_model_test.refresh()

                start_time = time.time()
                while custom_model_test.overall_status == 'in_progress':
                    if time.time() >= start_time + max_state_wait:
                        raise
                    time.sleep(1)
                    custom_model_test.refresh()

                return custom_model_test

    @classmethod
    def list(cls, custom_model_id):
        """List custom model tests.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            the id of the custom model

        Returns
        -------
        List[CustomModelTest]
            a list of custom model tests

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        payload = {"custom_model_id": custom_model_id}
        data = unpaginate(cls._path, payload, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, custom_model_test_id):
        """Get custom model test by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_test_id: str
            the id of the custom model test

        Returns
        -------
        CustomModelTest
            retrieved custom model test

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = "{}{}/".format(cls._path, custom_model_test_id)
        return cls.from_location(path)

    def get_log(self):
        """Get log of a custom model test.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = "{}{}/log/".format(self._path, self.id)
        return self._client.get(path).text

    def get_log_tail(self):
        """Get log tail of a custom model test.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = "{}{}/tail/".format(self._path, self.id)
        return self._client.get(path).text

    def cancel(self):
        """Cancel custom model test that is in progress.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = "{}{}/".format(self._path, self.id)
        self._client.delete(path)

    def refresh(self):
        """Update custom model test with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        path = "{}{}/".format(self._path, self.id)

        response = self._client.get(path)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))



import csv
from datetime import datetime
import json
import os
import pytest
import responses

from pandas import DataFrame
from pandas.testing import assert_frame_equal

from datarobot import DatetimeModel, enums

from tests.utils import request_body_to_json


def test_cannot_train_by_samplepct(datetime_model):
    with pytest.raises(NotImplementedError) as exc_info:
        datetime_model.train()
    assert 'use train_datetime instead' in str(exc_info.value)


def test_cannot_frozen_by_samplepct(datetime_model):
    with pytest.raises(NotImplementedError) as exc_info:
        datetime_model.request_frozen_model()
    assert 'use request_frozen_datetime_model instead' in str(exc_info.value)


def _validate_datetime_model(mod):
    assert mod.sample_pct is None

    training_info = mod.training_info
    expected_keys = {'holdout_training_start_date', 'holdout_training_duration',
                     'holdout_training_row_count', 'holdout_training_end_date',
                     'prediction_training_start_date', 'prediction_training_duration',
                     'prediction_training_row_count', 'prediction_training_end_date'}
    assert set(training_info.keys()) == expected_keys
    for key in training_info:
        if 'date' in key:
            value = training_info[key]
            assert isinstance(value, datetime)
            assert value.tzname() in ['UTC', 'Z']
    assert mod.data_selection_method in ['rowCount', 'duration',
                                         'selectedDateRange']
    assert mod.time_window_sample_pct is None or 0 < mod.time_window_sample_pct < 100
    assert mod.holdout_score is None
    assert mod.holdout_status == 'COMPLETED'
    train_dates = [mod.training_start_date, mod.training_end_date]
    for date in train_dates:
        if date is not None:
            assert isinstance(date, datetime)
            assert date.tzname() in ['UTC', 'Z']
    backtests = mod.backtests
    expected_keys = {'index', 'score', 'status', 'training_start_date', 'training_duration',
                     'training_row_count', 'training_end_date'}
    for bt in backtests:
        assert set(bt.keys()) == expected_keys
        for key in expected_keys:
            if 'date' in key and bt[key] is not None:
                date = bt[key]
                assert isinstance(date, datetime)
                assert date.tzname() in ['UTC', 'Z']
    fdw_keys = {
        'effective_feature_derivation_window_start', 'effective_feature_derivation_window_end',
        'forecast_window_start', 'forecast_window_end',
        'windows_basis_unit'
    }
    for key in fdw_keys:
        assert hasattr(mod, key)


@responses.activate
def test_retrieve_model(datetime_model_server_data, project_url):
    project_id = datetime_model_server_data['projectId']
    model_id = datetime_model_server_data['id']

    responses.add(responses.GET,
                  '{}datetimeModels/{}/'.format(project_url, datetime_model_server_data['id']),
                  json=datetime_model_server_data)

    mod = DatetimeModel.get(project_id, model_id)
    _validate_datetime_model(mod)


@responses.activate
def test_retrieve_model_with_startend_date(datetime_start_end_model_server_data, project_url):
    project_id = datetime_start_end_model_server_data['projectId']
    model_id = datetime_start_end_model_server_data['id']

    responses.add(responses.GET,
                  '{}datetimeModels/{}/'.format(project_url,
                                                datetime_start_end_model_server_data['id']),
                  json=datetime_start_end_model_server_data)

    mod = DatetimeModel.get(project_id, model_id)
    _validate_datetime_model(mod)


@responses.activate
def test_list_models(datetime_model_server_data, datetime_start_end_model_server_data,
                     datetime_project, project_url):
    json = {'previous': None, 'next': None, 'count': 2,
            'data': [datetime_model_server_data, datetime_start_end_model_server_data]}
    responses.add(responses.GET, '{}datetimeModels/'.format(project_url), json=json)
    mods = datetime_project.get_datetime_models()
    assert len(mods) == 2
    for mod in mods:
        _validate_datetime_model(mod)


@responses.activate
def test_train_from_project(datetime_project, project_url, datetime_model_job_server_data):
    model_job_loc = '{}modelJobs/1/'.format(project_url)
    responses.add(responses.POST, '{}datetimeModels/'.format(project_url),
                  body='', adding_headers={'Location': model_job_loc})
    responses.add(responses.GET, model_job_loc, json=datetime_model_job_server_data)

    mod_job = datetime_project.train_datetime('bp-id')

    assert mod_job.id == int(datetime_model_job_server_data['id'])
    assert mod_job.status == datetime_model_job_server_data['status']
    assert mod_job.processes == datetime_model_job_server_data['processes']
    assert mod_job.sample_pct is None


@responses.activate
def test_train_from_model(datetime_model, project_url, datetime_model_job_server_data):
    model_job_loc = '{}modelJobs/1/'.format(project_url)
    responses.add(responses.POST, '{}datetimeModels/'.format(project_url),
                  body='', adding_headers={'Location': model_job_loc})
    responses.add(responses.GET, model_job_loc, json=datetime_model_job_server_data)

    mod_job = datetime_model.train_datetime('bp-id')

    assert mod_job.id == int(datetime_model_job_server_data['id'])
    assert mod_job.status == datetime_model_job_server_data['status']
    assert mod_job.processes == datetime_model_job_server_data['processes']
    assert mod_job.sample_pct is None


@responses.activate
def train_request_frozen_datetime(datetime_model, project_url, datetime_model_job_server_data):
    model_job_loc = '{}modelJobs/{}/'.format(project_url, datetime_model_job_server_data['id'])
    responses.add(responses.POST, '{}frozenDatetimeModels/'.format(project_url),
                  body='', status=202, adding_headers={'Location': model_job_loc})
    responses.add(responses.GET, model_job_loc, json=datetime_model_job_server_data)

    mod_job = datetime_model.request_frozen_datetime_model(training_row_count=800)

    assert mod_job.id == int(datetime_model_job_server_data['id'])
    assert mod_job.status == datetime_model_job_server_data['status']
    assert mod_job.processes == datetime_model_job_server_data['processes']
    assert mod_job.sample_pct is None


@responses.activate
def train_request_frozen_datetime_with_dates(datetime_model, project_url,
                                             datetime_model_job_server_data):
    model_job_loc = '{}modelJobs/{}/'.format(project_url, datetime_model_job_server_data['id'])
    responses.add(responses.POST, '{}frozenDatetimeModels/'.format(project_url),
                  body='', status=202, adding_headers={'Location': model_job_loc})
    responses.add(responses.GET, model_job_loc, json=datetime_model_job_server_data)

    mod_job = datetime_model.request_frozen_datetime_model(training_start_date=datetime.now(),
                                                           training_end_date=datetime.now())

    assert mod_job.id == int(datetime_model_job_server_data['id'])
    assert mod_job.status == datetime_model_job_server_data['status']
    assert mod_job.processes == datetime_model_job_server_data['processes']
    assert mod_job.sample_pct is None


@responses.activate
def test_score_backtests(base_job_running_server_data, datetime_model, project_url):
    job_body = dict(base_job_running_server_data, jobType=enums.JOB_TYPE.MODEL)
    job_loc = '{}jobs/{}/'.format(project_url, job_body['id'])
    responses.add(responses.POST,
                  '{}datetimeModels/{}/backtests/'.format(project_url, datetime_model.id),
                  body='', adding_headers={'Location': job_loc}, status=202)
    responses.add(responses.GET, job_loc, json=job_body)

    job = datetime_model.score_backtests()
    assert job.id == int(job_body['id'])
    assert job.status == job_body['status']


def test_datetimemodel_dont_cross_validate(datetime_model):
    with pytest.raises(NotImplementedError):
        datetime_model.cross_validate()


@responses.activate
def test_compute_series_accuracy(base_job_running_server_data, datetime_model, project_url):
    job_body = dict(base_job_running_server_data, jobType=enums.JOB_TYPE.MODEL)
    job_loc = '{}jobs/{}/'.format(project_url, job_body['id'])
    responses.add(responses.POST,
                  '{}datetimeModels/{}/multiseriesScores/'.format(project_url, datetime_model.id),
                  body='', adding_headers={'Location': job_loc}, status=202)
    responses.add(responses.GET, job_loc, json=job_body)

    job = datetime_model.compute_series_accuracy()
    assert job.id == int(job_body['id'])
    assert job.status == job_body['status']


@responses.activate
def test_download_series_accuracy_as_csv(datetime_model, project_url, tmpdir):
    response_data = {
        u'querySeriesCount': 2, u'previous': None,
        u'data': [
            {u'multiseriesId': u'5c58975a211c0a1698093f25', u'validationScore': None,
             u'backtestingScore': None, u'rowCount': 365, u'multiseriesValues': [u'1'],
             u'holdoutScore': None, u'duration': u'P1Y0M0D'},
            {u'multiseriesId': u'5c58975a211c0a1698093f26', u'validationScore': None,
             u'backtestingScore': None, u'rowCount': 329, u'multiseriesValues': [u'2'],
             u'holdoutScore': None, u'duration': u'P0Y10M26D'}],
        u'totalSeriesCount': 2, u'next': None
    }
    url = '{}datetimeModels/{}/multiseriesScores/'.format(project_url, datetime_model.id)
    responses.add(responses.GET,
                  url,
                  body=json.dumps(response_data),
                  status=200)
    local_csv_filepath = tmpdir.join('series_accuracy.csv')
    datetime_model.download_series_accuracy_as_csv(
        local_csv_filepath.strpath,
        offset=1,
        limit=2,
        metric='asdf',
        multiseries_value='asdf',
        order_by='asdf',
        reverse=True
    )

    # Check correct params were used
    query_params = {'offset=1', 'limit=2', 'metric=asdf', 'multiseriesValue=asdf', 'orderBy=-asdf'}
    for query_param in query_params:
        assert query_param in responses.calls[0].request.url

    assert os.path.isfile(local_csv_filepath.strpath)
    with open(local_csv_filepath.strpath) as f:
        csv_output = csv.reader(f)
        # assert headers
        correct_header_list = ['multiseriesId', 'validationScore', 'backtestingScore',
                               'rowCount', 'multiseriesValues', 'holdoutScore', 'duration']
        assert sorted(next(csv_output)) == sorted(correct_header_list)
        for line in csv_output:
            assert len(line) == 7


@responses.activate
def test_download_series_accuracy_as_dataframe(datetime_model, project_url):
    response_data = {
        u'querySeriesCount': 2, u'previous': None,
        u'data': [
            {u'multiseriesId': u'5c58975a211c0a1698093f25', u'validationScore': None,
             u'backtestingScore': None, u'rowCount': 365, u'multiseriesValues': [u'1'],
             u'holdoutScore': None, u'duration': u'P1Y0M0D'},
            {u'multiseriesId': u'5c58975a211c0a1698093f26', u'validationScore': None,
             u'backtestingScore': None, u'rowCount': 329, u'multiseriesValues': [u'2'],
             u'holdoutScore': None, u'duration': u'P0Y10M26D'}],
        u'totalSeriesCount': 2, u'next': None
    }
    url = '{}datetimeModels/{}/multiseriesScores/'.format(project_url, datetime_model.id)
    responses.add(responses.GET,
                  url,
                  body=json.dumps(response_data),
                  status=200)
    df = datetime_model.get_series_accuracy_as_dataframe(
        offset=1,
        limit=2,
        metric='asdf',
        multiseries_value='asdf',
        order_by='asdf',
        reverse=True
    )

    # Check correct params were used
    query_params = {'offset=1', 'limit=2', 'metric=asdf', 'multiseriesValue=asdf', 'orderBy=-asdf'}
    for query_param in query_params:
        assert query_param in responses.calls[0].request.url

    assert_frame_equal(df, DataFrame(response_data['data']))


MONO_INC_FL_ID = '5a9975db962d74221ba38dc9'
MONO_DEC_FL_ID = '5a9975db962d74221ba38dca'


@pytest.mark.parametrize('featurelists', [
    (MONO_INC_FL_ID, MONO_DEC_FL_ID),
    (enums.MONOTONICITY_FEATURELIST_DEFAULT, enums.MONOTONICITY_FEATURELIST_DEFAULT),
    (None, None),
    (MONO_INC_FL_ID, None),
    (None, MONO_DEC_FL_ID),
    (MONO_INC_FL_ID, enums.MONOTONICITY_FEATURELIST_DEFAULT),
    (enums.MONOTONICITY_FEATURELIST_DEFAULT, MONO_DEC_FL_ID),
])
@responses.activate
def test_train_datetime_model_monotonicity(datetime_model,
                                           project_id,
                                           model_data,
                                           featurelists):
    responses.add(
        responses.POST,
        'https://host_name.com/projects/{}/datetimeModels/'.format(project_id),
        adding_headers={
            'Location': 'https://host_name.com/projects/{}/modelJobs/1/'.format(project_id)},
        body='',
        status=202
    )
    responses.add(
        responses.GET,
        'https://host_name.com/projects/{}/modelJobs/1/'.format(project_id),
        body=json.dumps({
            'blueprintId': datetime_model.blueprint_id,
            'featurelistId': datetime_model.featurelist_id,
            'id': 1,
            'isBlocked': False,
            'status': enums.QUEUE_STATUS.COMPLETED,
            'projectId': project_id,
            'modelType': datetime_model.model_type,
            'processes': datetime_model.processes,
        }),
        status=200
    )

    monotonic_increasing_featurelist_id, monotonic_decreasing_featurelist_id = featurelists

    datetime_model.train_datetime(
        monotonic_increasing_featurelist_id=monotonic_increasing_featurelist_id,
        monotonic_decreasing_featurelist_id=monotonic_decreasing_featurelist_id,
    )

    payload = request_body_to_json(responses.calls[0].request)

    if monotonic_increasing_featurelist_id is enums.MONOTONICITY_FEATURELIST_DEFAULT:
        assert 'monotonicIncreasingFeaturelistId' not in payload
    else:
        assert payload['monotonicIncreasingFeaturelistId'] == monotonic_increasing_featurelist_id

    if monotonic_decreasing_featurelist_id is enums.MONOTONICITY_FEATURELIST_DEFAULT:
        assert 'monotonicDecreasingFeaturelistId' not in payload
    else:
        assert payload['monotonicDecreasingFeaturelistId'] == monotonic_decreasing_featurelist_id


@responses.activate
def test_get_calculated_prediction_intervals(datetime_model, project_id):
    pred_int_url = 'https://host_name.com/projects/{}/models/{}/predictionIntervals/'.format(
        project_id, datetime_model.id
    )
    intervals = [80]
    responses.add(
        responses.GET,
        pred_int_url,
        body=json.dumps({'next': None, 'data': intervals, 'count': 1, 'previous': None}),
        status=200,
    )
    retrieved_intervals = datetime_model.get_calculated_prediction_intervals()
    assert intervals == retrieved_intervals


@responses.activate
def test_calculate_prediction_intervals(datetime_model, project_id):
    pred_int_url = 'https://host_name.com/projects/{}/models/{}/predictionIntervals/'.format(
        project_id, datetime_model.id
    )
    job_url = 'https://host_name.com/projects/{}/jobs/1/'.format(project_id)

    responses.add(
        responses.POST,
        pred_int_url,
        status=202,
        headers={'Location': job_url}
    )
    responses.add(
        responses.GET,
        job_url,
        status=303,
        content_type='application/json',
        body=json.dumps({
            'status': u'COMPLETED', 'model_id': datetime_model.id, 'is_blocked': False,
            'url': job_url, 'job_type': 'calculate_prediction_intervals', 'project_id': project_id,
            'id': 1
        }),
        headers={'Location': '{}result/'.format(job_url)}
    )

    job = datetime_model.calculate_prediction_intervals(80)
    assert job.job_type == 'calculate_prediction_intervals'
    assert request_body_to_json(responses.calls[0].request) == {'percentiles': [80]}



import json
from collections import OrderedDict

import dateutil
import pandas as pd
import pytest
import responses

from datarobot import Deployment
from datarobot.enums import ACCURACY_METRIC
from datarobot.models.deployment import Accuracy, AccuracyOverTime
from datarobot.utils import from_api


@pytest.fixture
def accuracy_response_data(deployment_data):
    return {
        'model_id': deployment_data['model']['id'],
        'period': {
            'start': '2019-08-01T00:00:00.000000Z',
            'end': '2019-08-10T00:00:00.000000Z'
        },
        'metrics': {
            'AUC': {'baselineValue': 1, 'value': 1.5, 'percentChange': 50},
            'LogLoss': {'baselineValue': None, 'value': None, 'percentChange': None}
        }
    }


@pytest.fixture
def deployment_get_accuracy_response(unittest_endpoint, deployment_data, accuracy_response_data):
    url = '{}/deployments/{}/accuracy/'.format(unittest_endpoint, deployment_data['id'])
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type='application/json',
        body=json.dumps(accuracy_response_data)
    )


@pytest.fixture
def accuracy_over_time_response_data(deployment_data):
    return {
        'model_id': deployment_data['model']['id'],
        'metric': 'LogLoss',
        'summary': {
            'period': {
                'start': '2019-08-01T00:00:00.000000Z',
                'end': '2019-08-04T00:00:00.000000Z'
            },
            'value': 1,
            'sampleSize': 100
        },
        'baseline': {
            'period': None,
            'value': None,
            'sampleSize': 0
        },
        'buckets': [
            {
                'period': {
                    'start': '2019-08-01T00:00:00.000000Z',
                    'end': '2019-08-02T00:00:00.000000Z'
                },
                'value': 0.22,
                'sampleSize': 30
            }, {
                'period': {
                    'start': '2019-08-02T00:00:00.000000Z',
                    'end': '2019-08-03T00:00:00.000000Z'
                },
                'value': 0.42,
                'sampleSize': 70
            }, {
                'period': {
                    'start': '2019-08-03T00:00:00.000000Z',
                    'end': '2019-08-04T00:00:00.000000Z'
                },
                'value': None,
                'sampleSize': 0
            }
        ]
    }


@pytest.fixture
def deployment_get_accuracy_over_time_response(unittest_endpoint, deployment_data,
                                               accuracy_over_time_response_data):
    url = '{}/deployments/{}/accuracyOverTime/'.format(unittest_endpoint, deployment_data['id'])
    responses.add(
        responses.GET,
        url,
        status=200,
        content_type='application/json',
        body=json.dumps(accuracy_over_time_response_data)
    )


def assert_bucket(actual, expected):
    assert sorted(actual.keys()) == sorted(expected.keys())

    actual_period = actual['period']
    if actual_period:
        assert actual['period'].keys() == expected['period'].keys()
        assert actual_period['start'] == dateutil.parser.parse(expected['period']['start'])
        assert actual_period['end'] == dateutil.parser.parse(expected['period']['end'])
    else:
        assert actual_period is None

    assert actual['value'] == expected['value']
    assert actual['sample_size'] == expected['sample_size']


@responses.activate
@pytest.mark.usefixtures('deployment_get_response', 'deployment_get_accuracy_response')
def test_accuracy(deployment_data, accuracy_response_data):
    accuracy_objs = [Accuracy.get(deployment_data['id']),
                     Deployment.get(deployment_data['id']).get_accuracy()]
    for accuracy in accuracy_objs:
        assert accuracy.model_id == accuracy_response_data['model_id']

        expected_start_date = dateutil.parser.parse(accuracy_response_data['period']['start'])
        expected_end_date = dateutil.parser.parse(accuracy_response_data['period']['end'])
        assert accuracy.period['start'] == expected_start_date
        assert accuracy.period['end'] == expected_end_date

        assert accuracy.metrics.keys() == accuracy_response_data['metrics'].keys()
        for metric_name, metric in accuracy.metrics.items():
            expected_metric = from_api(accuracy_response_data['metrics'][metric_name],
                                       keep_null_keys=True)
            assert metric == expected_metric
            assert accuracy[metric_name] == expected_metric['value']

        expected_metric_values = {name: metric['value']
                                  for name, metric in accuracy_response_data['metrics'].items()}
        assert accuracy.metric_values == expected_metric_values
        expected_metric_baseline = {name: metric['baselineValue']
                                    for name, metric in accuracy_response_data['metrics'].items()}
        assert accuracy.metric_baselines == expected_metric_baseline
        expected_percent_changes = {name: metric['percentChange']
                                    for name, metric in accuracy_response_data['metrics'].items()}
        assert accuracy.percent_changes == expected_percent_changes


@responses.activate
@pytest.mark.usefixtures('deployment_get_response', 'deployment_get_accuracy_over_time_response')
def test_accuracy_over_time(deployment_data, accuracy_over_time_response_data):
    accuracy_objs = [AccuracyOverTime.get(deployment_data['id']),
                     Deployment.get(deployment_data['id']).get_accuracy_over_time()]
    for accuracy_over_time in accuracy_objs:
        expected = from_api(accuracy_over_time_response_data, keep_null_keys=True)
        assert accuracy_over_time.metric == expected['metric']
        assert accuracy_over_time.model_id == expected['model_id']

        assert_bucket(accuracy_over_time.summary, expected['summary'])
        assert_bucket(accuracy_over_time.baseline, expected['baseline'])
        for index, bucket in enumerate(accuracy_over_time.buckets):
            assert_bucket(bucket, expected['buckets'][index])

        expected_bucket_values = [
            (dateutil.parser.parse(bucket['period']['start']), bucket['value'])
            for bucket in expected['buckets']]
        assert accuracy_over_time.bucket_values == OrderedDict(expected_bucket_values)

        expected_bucket_sample_sizes = [
            (dateutil.parser.parse(bucket['period']['start']), bucket['sample_size'])
            for bucket in expected['buckets']]
        assert accuracy_over_time.bucket_sample_sizes == OrderedDict(expected_bucket_sample_sizes)


@responses.activate
@pytest.mark.usefixtures('deployment_get_accuracy_over_time_response')
def test_accuracy_over_time_retrieve_as_dataframe(deployment_data,
                                                  accuracy_over_time_response_data):
    metric_names = [ACCURACY_METRIC.LOGLOSS, ACCURACY_METRIC.LOGLOSS]
    actual = AccuracyOverTime.get_as_dataframe(deployment_data['id'], metrics=metric_names)

    data = []
    start_times = []
    for bucket in accuracy_over_time_response_data['buckets']:
        start_time = dateutil.parser.parse(bucket['period']['start'])
        start_times.append(start_time)
        data.append([bucket['value']] * 2)
    expected = pd.DataFrame(data, index=start_times, columns=metric_names)

    assert actual.equals(expected)



import trafaret as t

from datarobot.errors import ClientError
from datarobot.models.lift_chart import LiftChartBinsTrafaret
from datarobot.utils.pagination import unpaginate
from ..api_object import APIObject
from ...utils import encode_utf8_if_py2
from .external_scores import DEFAULT_BATCH_SIZE


class ExternalLiftChart(APIObject):
    """ Lift chart for the model and prediction dataset with target or actual value column in
    unsupervised case.

    .. versionadded:: v2.21


    ``LiftChartBin`` is a dict containing the following:

        * ``actual`` (float) Sum of actual target values in bin
        * ``predicted`` (float) Sum of predicted target values in bin
        * ``bin_weight`` (float) The weight of the bin. For weighted projects, it is the sum of \
          the weights of the rows in the bin. For unweighted projects, it is the number of rows in \
          the bin.

    Attributes
    ----------
    dataset_id: str
        id of the prediction dataset with target or actual value column for unsupervised case
    bins: list of dict
        List of dicts with schema described as ``LiftChartBin`` above.

    """
    _path = 'projects/{project_id}/models/{model_id}/datasetLiftCharts/'

    _converter = t.Dict({
        t.Key('dataset_id'): t.String(),
    }).merge(LiftChartBinsTrafaret).ignore_extra('*')

    def __init__(self, dataset_id, bins):
        self.dataset_id = dataset_id
        self.bins = bins

    def __repr__(self):
        return encode_utf8_if_py2(
            u'ExternalLiftChart(dataset_id={}, bins={})'.format(self.dataset_id, self.bins))

    @classmethod
    def list(cls, project_id, model_id, dataset_id=None, offset=0, limit=100):
        """ Retrieve list of the lift charts for the model.

        Parameters
        ----------
        project_id: str
            id of the project
        model_id: str
            if specified, only lift chart for this model will be retrieved
        dataset_id: str, optional
            if specified, only lift chart for this dataset will be retrieved
        offset: int, optional
            this many results will be skipped, default: 0
        limit: int, optional
            at most this many results are returned, default: 100, max 1000.
            To return all results, specify 0

        Returns
        -------
            A list of :py:class:`ExternalLiftChart <datarobot.ExternalLiftChart>` objects
        """
        url = cls._path.format(project_id=project_id, model_id=model_id)
        params = {'limit': limit, 'offset': offset}
        if dataset_id:
            params['datasetId'] = dataset_id
        if limit == 0:  # unlimited results
            params['limit'] = DEFAULT_BATCH_SIZE
            return [cls.from_server_data(entry) for entry in unpaginate(url, params, cls._client)]
        r_data = cls._client.get(url, params=params).json()
        return [cls.from_server_data(item) for item in r_data['data']]

    @classmethod
    def get(cls, project_id, model_id, dataset_id):
        """ Retrieve lift chart for the model and prediction dataset.

        Parameters
        ----------
        project_id: str
            project id
        model_id: str
            model id
        dataset_id: str
            prediction dataset id with target or actual value column for unsupervised case

        Returns
        -------
            :py:class:`ExternalLiftChart <datarobot.ExternalLiftChart>` object

        """
        # always should return <=1 chart
        if dataset_id is None:
            raise ValueError('dataset_id must be specified')
        charts = cls.list(project_id, model_id, dataset_id=dataset_id)
        if not charts:
            raise ClientError('Requested lift chart does not exist.', 404)
        return charts[0]



import json

import dateutil
import pytest
import responses
import trafaret as t

from datarobot import Deployment
from datarobot.enums import (
    DEPLOYMENT_SERVICE_HEALTH_STATUS,
    DEPLOYMENT_MODEL_HEALTH_STATUS,
    DEPLOYMENT_ACCURACY_HEALTH_STATUS,
    DEPLOYMENT_EXECUTION_ENVIRONMENT_TYPE,
    DEPLOYMENT_IMPORTANCE,
)
from datarobot.models.deployment import DeploymentListFilters
from datarobot.utils import from_api
from tests.utils import request_body_to_json


def assert_deployment(deployment, deployment_data):
    def assert_health(health_from_client, health_from_api):
        assert health_from_client['status'] == health_from_client['status']
        if 'startDate' in health_from_api and health_from_api['startDate']:
            timestamp = health_from_api['startDate']
            assert health_from_client.get('start_date') == dateutil.parser.parse(timestamp)
        else:
            assert 'start_date' not in health_from_client
        if 'endDate' in health_from_api and health_from_api['endDate']:
            timestamp = health_from_api['endDate']
            assert health_from_client.get('end_date') == dateutil.parser.parse(timestamp)
        else:
            assert 'end_date' not in health_from_client

    assert deployment.id == deployment_data['id']
    assert deployment.label == deployment_data['label']
    assert deployment.description == deployment_data['description']

    assert deployment.model == from_api(deployment_data['model'])
    assert deployment.default_prediction_server == from_api(
        deployment_data['defaultPredictionServer'])
    assert deployment.capabilities == from_api(deployment_data['capabilities'])
    assert deployment.permissions == deployment_data['permissions']

    prediction_usage_from_api = deployment_data['predictionUsage']
    assert deployment.prediction_usage['daily_rates'] == prediction_usage_from_api['dailyRates']
    if prediction_usage_from_api['lastTimestamp']:
        assert deployment.prediction_usage['last_timestamp'] == dateutil.parser.parse(
            prediction_usage_from_api['lastTimestamp'])

    assert_health(deployment.service_health, deployment_data['serviceHealth'])
    assert_health(deployment.model_health, deployment_data['modelHealth'])
    assert_health(deployment.accuracy_health, deployment_data['accuracyHealth'])

    repr = 'Deployment({label})'.format(label=deployment_data['label'])
    assert str(deployment) == repr


@pytest.fixture()
def deployment_get_response(unittest_endpoint, deployments_data):
    for deployment_data in deployments_data:
        url = '{}/{}/{}/'.format(unittest_endpoint, 'deployments', deployment_data['id'])
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type='application/json',
            body=json.dumps(deployment_data)
        )


@responses.activate
@pytest.mark.usefixtures('deployment_get_response')
def test_get_deployment(deployments_data):
    for deployment_data in deployments_data:
        deployment = Deployment.get(deployment_data['id'])
        assert_deployment(deployment, deployment_data)


class TestDeploymentDelete(object):
    @pytest.fixture()
    def deployment_delete_response(self, unittest_endpoint, deployments_data):
        for deployment_data in deployments_data:
            url = '{}/{}/{}/'.format(unittest_endpoint, 'deployments', deployment_data['id'])
            responses.add(
                responses.DELETE,
                url,
                status=204
            )

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'deployment_delete_response')
    def test_delete_deployment(self, deployment_data):
        deployment = Deployment.get(deployment_data['id'])
        deployment.delete()

        assert responses.calls[1].request.method == 'DELETE'
        assert responses.calls[1].request.url.endswith('/deployments/5c6f7a9e9ca9b20017ff95a2/')


class TestDeploymentCreateFromLearningModel(object):
    @pytest.fixture
    def learning_model_create_response(self, unittest_endpoint, deployment_data):
        url = '{}/{}/{}/'.format(unittest_endpoint, 'deployments', 'fromLearningModel')
        responses.add(
            responses.POST,
            url,
            body=json.dumps({'id': deployment_data['id']}),
            status=200,
            content_type='application/json')

    @responses.activate
    @pytest.mark.usefixtures('learning_model_create_response', 'deployment_get_response')
    def test_create_minimum(self):
        Deployment.create_from_learning_model('5c76a543962d744efba25b85', 'loan is bad')

        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body['modelId'] == '5c76a543962d744efba25b85'
        assert create_body['label'] == 'loan is bad'
        assert 'description' not in create_body
        assert 'defaultPredictionServerId' not in create_body

    @responses.activate
    @pytest.mark.usefixtures('learning_model_create_response', 'deployment_get_response')
    def test_create_all_options(self):
        Deployment.create_from_learning_model(
            '5c76a543962d744efba25b85', 'loan is bad',
            'predict if a loan is gonna default', '5c0fcdb8962d74370dd0c38e')

        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body['modelId'] == '5c76a543962d744efba25b85'
        assert create_body['label'] == 'loan is bad'
        assert create_body['description'] == 'predict if a loan is gonna default'
        assert create_body['defaultPredictionServerId'] == '5c0fcdb8962d74370dd0c38e'


class TestDeploymentCreateFromCustomModelImage(object):
    def _make_model_package_get_response(self, get_model_package_url, custom_model_image_id, data):
        url = '{}?modelId={}'.format(get_model_package_url, custom_model_image_id)
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type='application/json',
            body=json.dumps({'count': 1, 'next': None, 'previous': None, 'data': data})
        )

    @pytest.fixture
    def get_model_package_url(self, unittest_endpoint):
        return '{}/modelPackages/'.format(unittest_endpoint)

    @pytest.fixture
    def create_model_package_url(self, unittest_endpoint):
        return '{}/modelPackages/fromCustomModelImage/'.format(unittest_endpoint)

    @pytest.fixture
    def get_deployment_url(self, unittest_endpoint):
        return '{}/deployments/'.format(unittest_endpoint)

    @pytest.fixture
    def create_deployment_url(self, unittest_endpoint):
        return '{}/deployments/fromModelPackage/'.format(unittest_endpoint)

    @pytest.fixture
    def get_status_url(self, unittest_endpoint):
        return '{}/status_url'.format(unittest_endpoint)

    @pytest.fixture
    def custom_model_image_id(self):
        return '5ea6aeeae8e8a40039714bde'

    @pytest.fixture
    def model_package_data(self):
        return {'id': '5eaaa09a40240308176cee55'}

    @pytest.fixture
    def prediction_server_id(self):
        return '5c0fcdb8962d74370dd0c38e'

    @pytest.fixture
    def model_package_get_response(
        self, get_model_package_url, custom_model_image_id, model_package_data
    ):
        self._make_model_package_get_response(
            get_model_package_url, custom_model_image_id, [model_package_data]
        )

    @pytest.fixture
    def model_package_get_response_none(self, get_model_package_url, custom_model_image_id):
        self._make_model_package_get_response(
            get_model_package_url, custom_model_image_id, []
        )

    @pytest.fixture
    def model_package_create_response(self, create_model_package_url, model_package_data):
        responses.add(
            responses.POST,
            create_model_package_url,
            body=json.dumps(model_package_data),
            status=200,
            content_type='application/json')

    @pytest.fixture
    def deployment_create_response(
        self, create_deployment_url, get_deployment_url, get_status_url, deployment_data
    ):
        responses.add(
            responses.POST,
            create_deployment_url,
            headers={'Location': get_status_url},
            body=json.dumps({'id': deployment_data['id']}),
            status=200,
            content_type='application/json')
        responses.add(
            responses.GET,
            get_status_url,
            headers={'Location': get_deployment_url},
            status=303)

    @pytest.fixture
    def _assert_deployment_creation(
            self,
            model_package_data,
            deployment_data,
            prediction_server_id,
            create_deployment_url,
            get_deployment_url,
            get_status_url,
    ):
        def _assert(
                create_deployment_req,
                get_status_req,
                get_deployment_req,
        ):
            assert create_deployment_req.method == 'POST'
            assert create_deployment_req.url == create_deployment_url
            create_body = request_body_to_json(create_deployment_req)
            assert set(create_body) == {
                'modelPackageId',
                'label',
                'description',
                'defaultPredictionServerId',
            }
            assert create_body['modelPackageId'] == model_package_data['id']
            assert create_body['label'] == 'My deployment'
            assert create_body['description'] == 'deployment description'
            assert create_body['defaultPredictionServerId'] == prediction_server_id

            assert get_status_req.method == 'GET'
            assert get_status_req.url == get_status_url

            assert get_deployment_req.method == 'GET'
            assert get_deployment_req.url == '{}{}/'.format(
                get_deployment_url, deployment_data['id']
            )

        return _assert

    @responses.activate
    @pytest.mark.usefixtures(
        'model_package_get_response',
        'deployment_create_response',
        'deployment_get_response'
    )
    def test_create_deployment_model_package_exists(
        self,
        custom_model_image_id,
        prediction_server_id,
        get_model_package_url,
        _assert_deployment_creation,
    ):
        Deployment.create_from_custom_model_image(
            custom_model_image_id,
            'My deployment',
            'deployment description',
            prediction_server_id,
        )

        assert len(responses.calls) == 4

        responses.calls[0].request.method == 'GET'
        assert responses.calls[0].request.url == '{}?modelId={}'.format(
            get_model_package_url, custom_model_image_id
        )

        create_deployment_req = responses.calls[1].request
        get_status_req = responses.calls[2].request
        get_deployment_req = responses.calls[3].request

        _assert_deployment_creation(
            create_deployment_req, get_status_req, get_deployment_req
        )

    @responses.activate
    @pytest.mark.usefixtures(
        'model_package_get_response_none',
        'model_package_create_response',
        'deployment_create_response',
        'deployment_get_response'
    )
    def test_create_deployment_model_package_doesnt_exist(
        self,
        custom_model_image_id,
        prediction_server_id,
        get_model_package_url,
        create_model_package_url,
        _assert_deployment_creation
    ):
        Deployment.create_from_custom_model_image(
            custom_model_image_id,
            'My deployment',
            'deployment description',
            prediction_server_id,
        )

        assert len(responses.calls) == 5

        # get model package
        assert responses.calls[0].request.method == 'GET'
        assert responses.calls[0].request.url == '{}?modelId={}'.format(
            get_model_package_url, custom_model_image_id
        )

        # create model package
        assert responses.calls[1].request.method == 'POST'
        assert responses.calls[1].request.url == create_model_package_url
        create_body = request_body_to_json(responses.calls[1].request)
        assert set(create_body) == {'customModelImageId'}
        assert create_body['customModelImageId'] == custom_model_image_id

        create_deployment_req = responses.calls[2].request
        get_status_req = responses.calls[3].request
        get_deployment_req = responses.calls[4].request

        _assert_deployment_creation(
            create_deployment_req, get_status_req, get_deployment_req
        )


class TestDeploymentList(object):
    @pytest.fixture()
    def deployments_list_response_data(self, deployments_data):
        return {
            'count': 1,
            'next': None,
            'previous': None,
            'data': deployments_data
        }

    @pytest.fixture()
    def deployment_list_response(self, unittest_endpoint, deployments_list_response_data):
        url = '{}/{}/'.format(unittest_endpoint, 'deployments')
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type='application/json',
            body=json.dumps(deployments_list_response_data)
        )

    @responses.activate
    @pytest.mark.usefixtures('deployment_list_response')
    def test_list_deployment(self, deployments_data):
        deployments_data = {item['id']: item for item in deployments_data}
        deployments = Deployment.list()
        assert len(deployments) == len(deployments_data)
        for deployment in deployments:
            assert_deployment(deployment, deployments_data[deployment.id])

    @responses.activate
    @pytest.mark.usefixtures('deployment_list_response')
    def test_sort(self):
        Deployment.list(order_by='-label')
        assert 'deployments/?orderBy=-label' in responses.calls[0].request.url

    @responses.activate
    @pytest.mark.usefixtures('deployment_list_response')
    def test_search(self):
        Deployment.list(search='readmitted')
        assert 'deployments/?search=readmitted' in responses.calls[0].request.url

    @responses.activate
    @pytest.mark.usefixtures('deployment_list_response')
    @pytest.mark.parametrize(
        'filter_name,filter_value,query_param_key,query_param_value', [
            ('role', 'OWNER', 'role', 'OWNER'),
            (
                    'service_health',
                    DEPLOYMENT_SERVICE_HEALTH_STATUS.PASSING,
                    'serviceHealth',
                    'passing'
            ),
            (
                    'service_health',
                    [DEPLOYMENT_SERVICE_HEALTH_STATUS.PASSING],
                    'serviceHealth',
                    'passing'
            ),
            (
                    'service_health',
                    [
                        DEPLOYMENT_SERVICE_HEALTH_STATUS.FAILING,
                        DEPLOYMENT_SERVICE_HEALTH_STATUS.WARNING
                    ],
                    'serviceHealth',
                    'failing,warning'
            ),
            ('model_health', DEPLOYMENT_MODEL_HEALTH_STATUS.PASSING, 'modelHealth', 'passing'),
            (
                    'model_health',
                    [DEPLOYMENT_MODEL_HEALTH_STATUS.FAILING],
                    'modelHealth',
                    'failing'
            ),
            (
                    'model_health',
                    [
                        DEPLOYMENT_MODEL_HEALTH_STATUS.PASSING,
                        DEPLOYMENT_MODEL_HEALTH_STATUS.UNKNOWN
                    ],
                    'modelHealth',
                    'passing,unknown'
            ),
            (
                    'accuracy_health',
                    DEPLOYMENT_ACCURACY_HEALTH_STATUS.UNAVAILABLE,
                    'accuracyHealth',
                    'unavailable'
            ),
            (
                    'accuracy_health',
                    [DEPLOYMENT_ACCURACY_HEALTH_STATUS.WARNING],
                    'accuracyHealth',
                    'warning'
            ),
            (
                    'accuracy_health',
                    [
                        DEPLOYMENT_ACCURACY_HEALTH_STATUS.FAILING,
                        DEPLOYMENT_ACCURACY_HEALTH_STATUS.PASSING
                    ],
                    'accuracyHealth',
                    'failing,passing'
            ),
            (
                    'importance',
                    [
                        DEPLOYMENT_IMPORTANCE.HIGH,
                        DEPLOYMENT_IMPORTANCE.LOW
                    ],
                    'importance',
                    'HIGH,LOW'
            ),
        ]
    )
    def test_filters(
            self, filter_name, filter_value, query_param_key, query_param_value
    ):
        filter_args = {filter_name: filter_value}
        filters = DeploymentListFilters(**filter_args)
        Deployment.list(filters=filters)

        expected_call = 'deployments/?{}={}'.format(
            query_param_key, query_param_value.replace(',', '%2C')
        )

        actual_call = responses.calls[0].request.url
        assert actual_call.endswith(expected_call), (
            'Expected called url to end with: {}\n Got: {}'.format(expected_call, actual_call)
        )

    @responses.activate
    @pytest.mark.usefixtures('deployment_list_response')
    def test_all_query_params(self):
        Deployment.list(
            order_by='-label',
            search='readmitted',
            filters=DeploymentListFilters(
                role='OWNER',
                service_health=DEPLOYMENT_SERVICE_HEALTH_STATUS.PASSING,
                model_health=DEPLOYMENT_MODEL_HEALTH_STATUS.WARNING,
                accuracy_health=DEPLOYMENT_ACCURACY_HEALTH_STATUS.FAILING,
                execution_environment_type=DEPLOYMENT_EXECUTION_ENVIRONMENT_TYPE.DATAROBOT,
            )
        )
        expected_args = [
            'orderBy=-label',
            'search=readmitted',
            'role=OWNER',
            'serviceHealth=passing',
            'modelHealth=warning',
            'accuracyHealth=failing',
            'executionEnvironmentType=datarobot',
        ]
        query_param_str = responses.calls[0].request.url.split('deployments/?')[1]
        query_params = query_param_str.split('&')

        assert len(expected_args) == len(query_params)

        for expected_arg in expected_args:
            assert expected_arg in query_params, (
                'Expected to find the query arg {}, but didn\'t.\n Got: {}'.format(
                    expected_arg, query_param_str
                )
            )


class TestDeploymentUpdate(object):
    @pytest.fixture()
    def deployment_update_response(self, unittest_endpoint, deployments_data):
        for deployment_data in deployments_data:
            url = '{}/{}/{}/'.format(unittest_endpoint, 'deployments', deployment_data['id'])
            responses.add(
                responses.PATCH,
                url,
                status=204
            )

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'deployment_update_response')
    def test_cannot_update_nothing(self, deployment_data):
        deployment = Deployment.get(deployment_data['id'])
        with pytest.raises(ValueError):
            deployment.update()

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'deployment_update_response')
    def test_update_label(self, deployment_data):
        deployment = Deployment.get(deployment_data['id'])
        deployment.update(label='new label')

        assert responses.calls[1].request.method == 'PATCH'
        assert responses.calls[1].request.url.endswith('/deployments/5c6f7a9e9ca9b20017ff95a2/')

        request_json = json.loads(responses.calls[1].request.body)
        assert request_json['label'] == 'new label'

    @responses.activate
    @pytest.mark.usefixtures('deployment_get_response', 'deployment_update_response')
    def test_update_description(self, deployment_data):
        deployment = Deployment.get(deployment_data['id'])
        deployment.update(description='new description')

        assert responses.calls[1].request.method == 'PATCH'
        assert responses.calls[1].request.url.endswith('/deployments/5c6f7a9e9ca9b20017ff95a2/')

        request_json = json.loads(responses.calls[1].request.body)
        assert request_json['description'] == 'new description'


class TestDeploymentListFilters(object):

    @pytest.mark.parametrize('filter_name,valid_value', [
        ('role', 'OWNER'),
        ('role', 'USER'),
        ('service_health', DEPLOYMENT_SERVICE_HEALTH_STATUS.PASSING),
        ('service_health', [DEPLOYMENT_SERVICE_HEALTH_STATUS.UNKNOWN]),
        ('service_health', DEPLOYMENT_SERVICE_HEALTH_STATUS.ALL),
        ('model_health', DEPLOYMENT_MODEL_HEALTH_STATUS.FAILING),
        ('model_health', [DEPLOYMENT_MODEL_HEALTH_STATUS.WARNING]),
        ('model_health', DEPLOYMENT_MODEL_HEALTH_STATUS.ALL),
        ('accuracy_health', DEPLOYMENT_ACCURACY_HEALTH_STATUS.UNAVAILABLE),
        ('accuracy_health', [DEPLOYMENT_ACCURACY_HEALTH_STATUS.UNKNOWN]),
        ('accuracy_health', DEPLOYMENT_ACCURACY_HEALTH_STATUS.ALL),
        ('execution_environment_type', DEPLOYMENT_EXECUTION_ENVIRONMENT_TYPE.DATAROBOT),
        ('execution_environment_type', DEPLOYMENT_EXECUTION_ENVIRONMENT_TYPE.ALL),
        ('importance', DEPLOYMENT_IMPORTANCE.MODERATE),
        ('importance', DEPLOYMENT_IMPORTANCE.ALL),
    ])
    def test_valid_filter_values(self, filter_name, valid_value):
        kwargs = {filter_name: valid_value}
        DeploymentListFilters(**kwargs)

    @pytest.mark.parametrize('filter_name,invalid_value', [
        ('role', 1234),
        ('service_health', 1234),
        ('model_health', 1234),
        ('accuracy_health', 1234),
        ('execution_environment_type', 1234),
        ('importance', 1234),
    ])
    def test_invalid_filter_values(self, filter_name, invalid_value):
        with pytest.raises(t.DataError):
            kwargs = {filter_name: invalid_value}
            DeploymentListFilters(**kwargs)



import json

import responses

from datarobot import ShapMatrixJob, enums
from datarobot.models.shap_matrix import ShapMatrix
from tests.test_job import mock_job_data


@responses.activate
def test_shap_matrix_job__refresh__ok():
    refresh_job_data = dict(mock_job_data, status=enums.QUEUE_STATUS.ERROR)
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=200,
        body=json.dumps(refresh_job_data),
        content_type='application/json'
    )

    job = ShapMatrixJob(mock_job_data, model_id='model-id', dataset_id='dataset-id')
    job.refresh()

    assert job.status == enums.QUEUE_STATUS.ERROR


@responses.activate
def test_shap_matrix_job__wait_for_completion__ok():
    resource_url = 'https://host_name.com/projects/p-id/shapMatrices/shap-matrix-id'
    refresh_job_data = dict(mock_job_data, status=enums.QUEUE_STATUS.COMPLETED)
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=303,
        body=json.dumps(refresh_job_data),
        content_type='application/json',
        adding_headers={'Location': resource_url},
    )

    job = ShapMatrixJob(mock_job_data, model_id='model-id', dataset_id='dataset-id')
    job.wait_for_completion(max_wait=3)

    assert job.status == enums.QUEUE_STATUS.COMPLETED


@responses.activate
def test_get_completed_shap_matrix_with_shap_matrix_job(client):
    job_id = '1'
    project_id = 'p-id'
    shap_matrix_id = 'shap-matrix-id'
    data_path = "projects/{project_id}/shapMatrices/{shap_matrix_id}/".format(
        project_id=project_id, shap_matrix_id=shap_matrix_id)
    get_shap_matrix_url = 'https://host_name.com/{data_path}'.format(data_path=data_path)
    mock_generic_job_data = {
        'status': enums.QUEUE_STATUS.COMPLETED,
        'url': 'https://host_name.com/projects/p-id/predictJobs/1/',
        'id': job_id,
        'jobType': 'a job',
        'isBlocked': False,
        'projectId': project_id
    }

    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=303,
        body=json.dumps(dict(mock_generic_job_data)),
        content_type='application/json',
        adding_headers={'Location': get_shap_matrix_url}
    )

    job = ShapMatrixJob.get(project_id, job_id)

    data = job.get_result()
    assert job.status == 'COMPLETED'

    assert isinstance(data, ShapMatrix)
    assert data.project_id == project_id
    assert data.id == shap_matrix_id
    assert data.model_id is None
    assert data.dataset_id is None



import json

import pytest
import responses
from datarobot.errors import ClientError

from datarobot import Model

from datarobot.models.lift_chart import LiftChart


@pytest.fixture
def lift_chart_validation_data(model_id, lift_chart_bins_data):
    return {
        'source': 'validation',
        'bins': lift_chart_bins_data,
        'sourceModelId': model_id
    }


@pytest.fixture
def multiclass_lift_chart_validation_data(model_id, lift_chart_bins_data):
    return {
        'source': 'validation',
        'classBins': [
            {
                'targetClass': 'classA',
                'bins': lift_chart_bins_data
            },
            {
                'targetClass': 'classB',
                'bins': lift_chart_bins_data
            },
            {
                'targetClass': 'classC',
                'bins': lift_chart_bins_data
            }
        ]
    }


@pytest.fixture
def lift_chart_parent_model_validation_data(parent_model_id, lift_chart_bins_data):
    return {
        'source': 'validation',
        'bins': lift_chart_bins_data,
        'sourceModelId': parent_model_id
    }


@pytest.fixture
def lift_chart_parent_model_holdout_data(parent_model_id, lift_chart_bins_data):
    return {
        'source': 'holdout',
        'bins': lift_chart_bins_data,
        'sourceModelId': parent_model_id
    }


@pytest.fixture
def expected_bin_data(lift_chart_validation_data):
    expected_bins = [dict(bin_data) for bin_data in lift_chart_validation_data['bins']]
    for expected in expected_bins:
        weight = expected.pop('binWeight')
        expected['bin_weight'] = weight
    return expected_bins


def test_instantiation(lift_chart_validation_data, expected_bin_data):
    lc = LiftChart.from_server_data(lift_chart_validation_data)

    assert lc.source == lift_chart_validation_data['source']
    assert lc.bins == expected_bin_data


def test_future_proof(lift_chart_validation_data):
    data_with_future_keys = dict(lift_chart_validation_data, new_key='some future lift data')
    data_with_future_keys['bins'][0]['new_key'] = 'some future bin data'
    LiftChart.from_server_data(data_with_future_keys)


@pytest.fixture
def lift_chart_validation_data_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/liftChart/validation/'.format(
        project_id,
        model_id
    )


@pytest.fixture
def multiclass_lift_chart_validation_data_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/multiclassLiftChart/validation/'.format(
        project_id,
        model_id
    )


@pytest.fixture
def lift_chart_parent_model_validation_data_url(project_id, parent_model_id):
    return 'https://host_name.com/projects/{}/models/{}/liftChart/validation/'.format(
        project_id,
        parent_model_id
    )


@pytest.fixture
def lift_chart_list_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/liftChart/'.format(
        project_id,
        model_id
    )


@pytest.fixture
def lift_chart_parent_list_url(project_id, parent_model_id):
    return 'https://host_name.com/projects/{}/models/{}/liftChart/'.format(
        project_id,
        parent_model_id
    )


@pytest.fixture
def frozen_model_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/frozenModels/{}/'.format(
        project_id,
        model_id
    )


@responses.activate
def test_get_validation_lift_chart(lift_chart_validation_data,
                                   expected_bin_data,
                                   lift_chart_validation_data_url,
                                   project_id, model_id):
    responses.add(
        responses.GET,
        lift_chart_validation_data_url,
        status=200,
        content_type='application/json',
        body=json.dumps(lift_chart_validation_data)
    )
    model = Model(id=model_id, project_id=project_id)
    lc = model.get_lift_chart('validation')

    assert lc.source == lift_chart_validation_data['source']
    assert lc.bins == expected_bin_data
    assert lc.source_model_id == model_id


@responses.activate
def test_get_validation_multiclass_lift_chart(multiclass_lift_chart_validation_data,
                                              expected_bin_data,
                                              multiclass_lift_chart_validation_data_url,
                                              project_id, model_id):
    responses.add(
        responses.GET,
        multiclass_lift_chart_validation_data_url,
        status=200,
        content_type='application/json',
        body=json.dumps(multiclass_lift_chart_validation_data)
    )
    model = Model(id=model_id, project_id=project_id)
    lcs = model.get_multiclass_lift_chart('validation')
    # 3-class multiclass contains 3 records for each class
    assert len(lcs) == 3
    for lc in lcs:
        assert lc.source == multiclass_lift_chart_validation_data['source']
        assert lc.bins == expected_bin_data
        assert lc.source_model_id == model_id


@responses.activate
def test_get_frozen_validation_lift_chart_no_fallback(lift_chart_validation_data_url,
                                                      project_id, model_id):
    responses.add(
        responses.GET,
        lift_chart_validation_data_url,
        status=404
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    with pytest.raises(ClientError):
        model.get_lift_chart('validation')


@responses.activate
@pytest.mark.usefixtures('known_warning')
def test_get_frozen_validation_lift_chart_with_fallback(
        lift_chart_parent_model_validation_data,
        expected_bin_data,
        lift_chart_validation_data_url,
        lift_chart_parent_model_validation_data_url,
        frozen_model_url,
        project_id, model_id, parent_model_id,
        frozen_json):

    responses.add(
        responses.GET,
        lift_chart_validation_data_url,
        status=404
    )
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type='application/json'
    )
    responses.add(
        responses.GET,
        lift_chart_parent_model_validation_data_url,
        status=200,
        content_type='application/json',
        body=json.dumps(lift_chart_parent_model_validation_data)
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    lc = model.get_lift_chart('validation', fallback_to_parent_insights=True)

    assert lc.source == lift_chart_parent_model_validation_data['source']
    assert lc.bins == expected_bin_data
    assert lc.source_model_id == parent_model_id


@responses.activate
def test_get_all_lift_charts(lift_chart_validation_data,
                             expected_bin_data,
                             lift_chart_list_url,
                             project_id, model_id):
    responses.add(
        responses.GET,
        lift_chart_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [lift_chart_validation_data]})
    )
    model = Model(id=model_id, project_id=project_id)
    lc_list = model.get_all_lift_charts()

    assert len(lc_list) == 1
    assert lc_list[0].source == lift_chart_validation_data['source']
    assert lc_list[0].bins == expected_bin_data
    assert lc_list[0].source_model_id == model_id


@responses.activate
def test_get_frozen_all_lift_charts_no_fallback(lift_chart_validation_data,
                                                expected_bin_data,
                                                lift_chart_list_url,
                                                project_id, model_id):
    responses.add(
        responses.GET,
        lift_chart_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [lift_chart_validation_data]})
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    lc_list = model.get_all_lift_charts()

    assert len(lc_list) == 1
    assert lc_list[0].source == lift_chart_validation_data['source']
    assert lc_list[0].bins == expected_bin_data
    assert lc_list[0].source_model_id == model_id


@responses.activate
@pytest.mark.usefixtures('known_warning')
def test_get_frozen_all_lift_charts_with_fallback(lift_chart_validation_data,
                                                  lift_chart_parent_model_validation_data,
                                                  lift_chart_parent_model_holdout_data,
                                                  expected_bin_data,
                                                  lift_chart_list_url,
                                                  lift_chart_parent_list_url,
                                                  frozen_model_url,
                                                  project_id, model_id, parent_model_id,
                                                  frozen_json):
    responses.add(
        responses.GET,
        lift_chart_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [lift_chart_validation_data]})
    )
    responses.add(
        responses.GET,
        frozen_model_url,
        body=frozen_json,
        status=200,
        content_type='application/json'
    )
    responses.add(
        responses.GET,
        lift_chart_parent_list_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'charts': [lift_chart_parent_model_validation_data,
                                    lift_chart_parent_model_holdout_data]})
    )
    model = Model(id=model_id, project_id=project_id, is_frozen=True)
    lc_list = model.get_all_lift_charts(fallback_to_parent_insights=True)

    assert len(lc_list) == 2
    assert lc_list[0].source == lift_chart_validation_data['source']
    assert lc_list[0].bins == expected_bin_data
    assert lc_list[0].source_model_id == model_id

    assert lc_list[1].source == lift_chart_parent_model_holdout_data['source']
    assert lc_list[1].bins == expected_bin_data
    assert lc_list[1].source_model_id == parent_model_id



import json
from datetime import datetime

import pandas as pd
import pytest
import responses
from dateutil.tz import tzutc

from datarobot import Predictions, errors
from datarobot.utils import parse_time
from tests.utils import add_response


@pytest.fixture
def sample_prediction():
    return {
        u'positiveClass': None,
        u'task': u'Regression',
        u'predictions': [
            {u'positiveProbability': None, u'prediction': 32.0,
             u'rowId': 0},
            {u'positiveProbability': None, u'prediction': 100.0,
             u'rowId': 1},
            {u'positiveProbability': None, u'prediction': 212.0,
             u'rowId': 2}
        ]
    }


@pytest.fixture
def predictions_regression_with_shap_explanations():
    return {
        'shapBaseValue': [31108.28125],
        'task': 'Regression',
        'positiveClass': None,
        'predictions': [
            {
                'positiveProbability': None,
                'predictionExplanationMetadata': {'shapRemainingTotal': -6969.941378326501},
                'prediction': 27229.166666666668,
                'rowId': 0,
                'predictionExplanations': [
                    {
                        'featureValue': '2004',
                        'strength': 3090.8262156678165,
                        'feature': 'YearMade',
                        'label': 'SalePrice',
                    },
                    {
                        'featureValue': '999089',
                        'strength': -2788.4580943655205,
                        'feature': 'MachineID',
                        'label': 'SalePrice',
                    },
                    {
                        'featureValue': '2006',
                        'strength': -2286.181442643504,
                        'feature': 'saledate (Year)',
                        'label': 'SalePrice',
                    },
                ],
            },
            {
                'positiveProbability': None,
                'predictionExplanationMetadata': {'shapRemainingTotal': 13439.981135022792},
                'prediction': 60583.333333333336,
                'rowId': 1,
                'predictionExplanations': [
                    {
                        'featureValue': '117657',
                        'strength': 16035.069717943523,
                        'feature': 'MachineID',
                        'label': 'SalePrice',
                    },
                    {
                        'featureValue': '77',
                        'strength': 10105.427194858292,
                        'feature': 'ModelID',
                        'label': 'SalePrice',
                    },
                    {
                        'featureValue': 'Medium',
                        'strength': 3432.899846833262,
                        'feature': 'ProductSize',
                        'label': 'SalePrice',
                    },
                ],
            },
        ],
    }


@pytest.fixture
def predictions_binary_with_shap_explanations():
    return {
        'shapBaseValue': [0.39640626311302185],
        'task': 'Binary',
        'positiveClass': 1.0,
        'predictions': [
            {
                'positiveProbability': 0.311178247734139,
                'prediction': 0.0,
                'predictionExplanations': [
                    {
                        'featureValue': '9',
                        'strength': 0.05346482340175079,
                        'feature': 'number_diagnoses',
                        'label': 1.0,
                    },
                    {
                        'featureValue': '0',
                        'strength': -0.049024358641319835,
                        'feature': 'number_inpatient',
                        'label': 1.0,
                    },
                    {
                        'featureValue': 'Surgery-Neuro',
                        'strength': -0.03165299590594699,
                        'feature': 'medical_specialty',
                        'label': 1.0,
                    },
                ],
                'rowId': 0,
                'predictionValues': [
                    {'value': 0.3111782477000000, 'label': 1.0},
                    {'value': 0.6888217522658611, 'label': 0.0},
                ],
                # To test the case with empty metadata from API.
                'predictionExplanationMetadata': {'shapRemainingTotal': None},
                'predictionThreshold': 0.5,
            },
            {
                'positiveProbability': 0.2829861111111111,
                'prediction': 0.0,
                'predictionExplanations': [
                    {
                        'featureValue': '0',
                        'strength': -0.05744066866980976,
                        'feature': 'number_inpatient',
                        'label': 1.0,
                    },
                    {
                        'featureValue': '6',
                        'strength': -0.04089034311313498,
                        'feature': 'number_diagnoses',
                        'label': 1.0,
                    },
                    {
                        'featureValue': '[50-75)',
                        'strength': 0.03388593271674542,
                        'feature': 'weight',
                        'label': 1.0,
                    },
                ],
                'rowId': 1,
                'predictionValues': [
                    {'value': 0.2829861111111111, 'label': 1.0},
                    {'value': 0.7170138888888888, 'label': 0.0},
                ],
                'predictionExplanationMetadata': {'shapRemainingTotal': -0.00887273281105585},
                'predictionThreshold': 0.5,
            },
        ],
    }


@pytest.fixture
def predictions_with_shap_exp_csv():
    return """row_id,prediction,Explanation_1_feature_name,Explanation_1_feature_value,Explanation_1_strength,shap_remaining_total,shap_base_value
    0,27229.166666666668,YearMade,2004,3090.8262156678165,-6969.941378326501,31108.28125
    1,60583.333333333336,MachineID,117657,16035.069717943523,13439.981135022792,31108.28125
    2,11800.0,fiSecondaryDesc,nan,-14385.515413569377,-4922.765764714323,31108.28125
    3,17453.125,fiSecondaryDesc,nan,-7614.76113549681,-6040.3950427868795,31108.28125
    4,11539.473684210527,Hydraulics_Flow,Standard,-14743.391828540229,-4825.41571693096,31108.28125
    5,22555.555555555555,MachineID,1001274,-6044.702269899164,-2508.024220884521,31108.28125
    6,60583.333333333336,MachineID,772701,20612.841554550116,8862.20929841619,31108.28125
    7,31200.0,MachineID,902002,13999.795343335558,-13908.076521619252,31108.28125
    8,67863.63636363637,YearMade,2008,30190.432159295826,6564.919474920468,31108.28125
    9,21176.470588235294,ProductSize,Large,7127.206390025408,-17059.016865184112,31108.28125
    """


@pytest.fixture
def predictions_metadata_with_shap_exp(prediction_id, project_id, model_id, dataset_id):
    url = 'http://localhost/api/v2/projects/{}/predictions/{}/'.format(project_id, prediction_id)
    return {
        'url': url,
        'id': prediction_id,
        'projectId': project_id,
        'modelId': model_id,
        'predictionDatasetId': dataset_id,
        'actualValueColumn': None,
        'forecastPoint': None,
        'includesPredictionIntervals': False,
        'predictionIntervalsSize': None,
        'predictionsStartDate': None,
        'predictionsEndDate': None,
        'explanationAlgorithm': 'shap',
        'maxExplanations': 1,
        'shapWarnings': {
            'mismatchRowCount': 1,
            'maxNormalizedMismatch': 0.01,
        },

    }


@pytest.fixture
def sample_prediction_metadata(prediction_id):
    project_id = '59d92fe8962d7464ca8c6bd6'
    url = 'http://localhost/api/v2/projects/{}/predictions/{}/'.format(
        project_id,
        prediction_id)

    return {
        'projectId': project_id,
        'id': prediction_id,
        'predictionDatasetId': 'a_dataset_id',
        'modelId': 'a_model_id',
        'includesPredictionIntervals': False,
        'predictionIntervalsSize': None,
        'url': url,
        'forecastPoint': '2017-01-01T15:00:00Z',
        'predictionsStartDate': None,
        'predictionsEndDate': None,
    }


@pytest.fixture
def prediction_id():
    return '59d92fe8962d7464ca8c6bd6'


@pytest.fixture
def model_id():
    return "modelid"


@pytest.fixture
def dataset_id():
    return "datasetid"


@pytest.fixture
def predictions_url(project_id, prediction_id):
    return 'https://host_name.com/projects/{}/predictions/{}/'.format(project_id, prediction_id)


@pytest.fixture
def predictions_metadata_url(project_id):
    return 'https://host_name.com/projects/{}/predictionsMetadata/'.format(project_id)


@pytest.fixture
def predictions_metadata_by_id_url(predictions_metadata_url, prediction_id):
    return predictions_metadata_url + '{}/'.format(prediction_id)


@responses.activate
def test_get__fetches_metadata(project_id, prediction_id, sample_prediction_metadata):
    metadata_url = 'https://host_name.com/projects/{}/predictionsMetadata/{}/'.format(
        project_id,
        prediction_id,
    )
    responses.add(
        responses.GET,
        metadata_url,
        body=json.dumps(sample_prediction_metadata),
        status=200,
        content_type='application/json',
    )

    preds = Predictions.get(project_id, prediction_id)

    assert preds.project_id == project_id
    assert preds.prediction_id == prediction_id
    assert preds.model_id == sample_prediction_metadata['modelId']
    assert preds.dataset_id == sample_prediction_metadata['predictionDatasetId']
    assert preds.includes_prediction_intervals == (
            sample_prediction_metadata['includesPredictionIntervals']
    )
    assert preds.forecast_point == parse_time(sample_prediction_metadata['forecastPoint'])
    assert preds.predictions_start_date == parse_time(
        sample_prediction_metadata['predictionsStartDate']
    )
    assert preds.predictions_end_date == parse_time(
        sample_prediction_metadata['predictionsEndDate']
    )
    assert preds.explanation_algorithm is None
    assert preds.max_explanations is None
    assert preds.shap_warnings is None


@responses.activate
def test_get_predictions_with_shap_explanations(
        project_id,
        prediction_id,
        predictions_metadata_with_shap_exp,
        predictions_metadata_by_id_url,
):
    mocked_response = predictions_metadata_with_shap_exp
    add_response(url=predictions_metadata_by_id_url, body=mocked_response)

    predictions = Predictions.get(project_id, prediction_id)

    assert predictions.project_id == project_id
    assert predictions.prediction_id == prediction_id
    assert predictions.model_id == mocked_response['modelId']
    assert predictions.dataset_id == mocked_response['predictionDatasetId']
    assert predictions.actual_value_column is None
    assert predictions.forecast_point is None
    assert not predictions.includes_prediction_intervals
    assert predictions.predictions_start_date is None
    assert predictions.predictions_end_date is None
    assert predictions.explanation_algorithm == mocked_response['explanationAlgorithm']
    assert predictions.max_explanations == mocked_response['maxExplanations']
    assert predictions.shap_warnings is not None
    assert predictions.shap_warnings[
               'mismatch_row_count'
           ] == mocked_response['shapWarnings']['mismatchRowCount']
    assert predictions.shap_warnings[
               'max_normalized_mismatch'
           ] == mocked_response['shapWarnings']['maxNormalizedMismatch']


@responses.activate
def test_list__ok(project_id, model_id, prediction_id, dataset_id, predictions_metadata_url):
    stub = [
        {
            'url': 'http://localhost/api/v2/projects/{}/predictions/{}/'.format(project_id,
                                                                                prediction_id),
            'id': prediction_id,
            'modelId': model_id,
            'predictionDatasetId': dataset_id,
            'includesPredictionIntervals': False,
            'forecastPoint': '2017-01-01T15:00:00Z',
        },
        {
            'url': 'http://localhost/api/v2/projects/{}/predictions/{}/'.format(project_id,
                                                                                prediction_id),
            'id': prediction_id,
            'modelId': model_id,
            'predictionDatasetId': dataset_id,
            'includesPredictionIntervals': False,
            'predictionsStartDate': '2019-05-19T00:00:00.000000Z',
            'predictionsEndDate': '2019-05-20T00:00:00.000000Z',
            'actualValueColumn': 'actuals',
        },
        {
            'url': 'http://localhost/api/v2/projects/{}/predictions/{}/'.format(
                project_id, prediction_id
            ),
            'id': prediction_id,
            'projectId': project_id,
            'modelId': model_id,
            'predictionDatasetId': dataset_id,
            'includesPredictionIntervals': False,
            'predictionsStartDate': None,
            'predictionsEndDate': None,
            'predictionIntervalsSize': None,
            'actualValueColumn': None,
            'forecastPoint': None,
            'explanationAlgorithm': 'shap',
            'maxExplanations': 1,
            'shapWarnings': None,
        },
    ]
    add_response(url=predictions_metadata_url, body={'data': stub})

    predictions_list = Predictions.list(project_id)

    assert len(predictions_list) == 3
    prediction = predictions_list[0]
    assert prediction.project_id == project_id
    assert prediction.model_id == model_id
    assert prediction.dataset_id == dataset_id
    assert prediction.prediction_id == prediction_id
    assert not prediction.includes_prediction_intervals
    assert prediction.prediction_intervals_size is None
    assert prediction.forecast_point == datetime(2017, 1, 1, 15, tzinfo=tzutc())
    assert prediction.predictions_start_date is None
    assert prediction.predictions_end_date is None
    assert prediction.explanation_algorithm is None
    assert prediction.max_explanations is None
    assert prediction.shap_warnings is None
    bulk_pred_metadata = predictions_list[1]
    assert bulk_pred_metadata.forecast_point is None
    assert bulk_pred_metadata.predictions_start_date == datetime(2019, 5, 19, tzinfo=tzutc())
    assert bulk_pred_metadata.predictions_end_date == datetime(2019, 5, 20, tzinfo=tzutc())
    assert bulk_pred_metadata.actual_value_column == 'actuals'
    assert bulk_pred_metadata.explanation_algorithm is None
    assert bulk_pred_metadata.max_explanations is None
    assert bulk_pred_metadata.shap_warnings is None
    predictions_with_shap_exp = predictions_list[2]
    assert predictions_with_shap_exp.project_id == project_id
    assert predictions_with_shap_exp.model_id == model_id
    assert predictions_with_shap_exp.dataset_id == dataset_id
    assert predictions_with_shap_exp.prediction_id == prediction_id
    assert not predictions_with_shap_exp.includes_prediction_intervals
    assert predictions_with_shap_exp.prediction_intervals_size is None
    assert predictions_with_shap_exp.forecast_point is None
    assert predictions_with_shap_exp.predictions_start_date is None
    assert predictions_with_shap_exp.predictions_end_date is None
    assert predictions_with_shap_exp.explanation_algorithm == 'shap'
    assert predictions_with_shap_exp.max_explanations == 1
    assert predictions_with_shap_exp.shap_warnings is None


@responses.activate
def test_get_all_as_dataframe__ok(
        project_id,
        prediction_id,
        project_with_target_json,
        sample_prediction,
        sample_prediction_metadata,
        predictions_url,
        predictions_metadata_by_id_url,
):
    add_response(predictions_url, sample_prediction)
    add_response(predictions_metadata_by_id_url, sample_prediction_metadata)

    obj = Predictions.get(project_id, prediction_id)
    data_frame = obj.get_all_as_dataframe()

    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (len(sample_prediction['predictions']), 2)
    assert list(data_frame.columns) == [
        'prediction',
        'row_id',
    ]


@responses.activate
def test_get_all_as_dataframe_regression_json_serializer(
        project_id,
        prediction_id,
        predictions_metadata_by_id_url,
        predictions_metadata_with_shap_exp,
        predictions_url,
        predictions_regression_with_shap_explanations,
):
    add_response(predictions_metadata_by_id_url, predictions_metadata_with_shap_exp)
    add_response(predictions_url, predictions_regression_with_shap_explanations)

    predictions = Predictions.get(project_id, prediction_id)
    data_frame = predictions.get_all_as_dataframe(serializer='json')
    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (2, 13)
    assert 'row_id' in data_frame.columns
    assert 'prediction' in data_frame.columns
    assert 'shap_remaining_total' in data_frame.columns
    assert 'shap_base_value' in data_frame.columns
    explanation_columns = [column for column in data_frame.columns if 'Explanation_' in column]
    assert len(explanation_columns) == 9
    feature_name_columns = [column for column in explanation_columns if 'feature_name' in column]
    assert len(feature_name_columns) == 3
    feature_value_columns = [column for column in explanation_columns if 'feature_value' in column]
    assert len(feature_value_columns) == 3
    strength_columns = [column for column in explanation_columns if 'strength' in column]
    assert len(strength_columns) == 3


@responses.activate
def test_get_all_as_dataframe_binary_json_serializer(
        project_id,
        prediction_id,
        predictions_metadata_by_id_url,
        predictions_metadata_with_shap_exp,
        predictions_url,
        predictions_binary_with_shap_explanations,
):
    add_response(predictions_metadata_by_id_url, predictions_metadata_with_shap_exp)
    add_response(predictions_url, predictions_binary_with_shap_explanations)

    predictions = Predictions.get(project_id, prediction_id)
    data_frame = predictions.get_all_as_dataframe(serializer='json')
    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (2, 18)
    assert 'row_id' in data_frame.columns
    assert 'prediction' in data_frame.columns
    assert 'shap_remaining_total' in data_frame.columns
    assert 'shap_base_value' in data_frame.columns
    assert 'explained_class' in data_frame.columns
    explanation_columns = [column for column in data_frame.columns if 'Explanation_' in column]
    assert len(explanation_columns) == 9
    feature_name_columns = [column for column in explanation_columns if 'feature_name' in column]
    assert len(feature_name_columns) == 3
    feature_value_columns = [column for column in explanation_columns if 'feature_value' in column]
    assert len(feature_value_columns) == 3
    strength_columns = [column for column in explanation_columns if 'strength' in column]
    assert len(strength_columns) == 3


@responses.activate
def test_get_all_as_dataframe_csv_serializer(
        project_id,
        prediction_id,
        predictions_url,
        predictions_metadata_by_id_url,
        predictions_metadata_with_shap_exp,
        predictions_with_shap_exp_csv,
):
    add_response(url=predictions_metadata_by_id_url, body=predictions_metadata_with_shap_exp)
    add_response(url=predictions_url, body=predictions_with_shap_exp_csv, content_type='text/csv')

    predictions = Predictions.get(project_id, prediction_id)
    data_frame = predictions.get_all_as_dataframe(serializer='csv')
    assert isinstance(data_frame, pd.DataFrame)
    assert data_frame.shape == (10, 7)
    assert 'row_id' in data_frame.columns
    assert 'prediction' in data_frame.columns
    assert 'shap_remaining_total' in data_frame.columns
    assert 'shap_base_value' in data_frame.columns
    explanation_columns = [column for column in data_frame.columns if 'Explanation_1' in column]
    assert len(explanation_columns) == 3
    assert any(['feature_name' in column for column in explanation_columns])
    assert any(['feature_value' in column for column in explanation_columns])
    assert any(['strength' in column for column in explanation_columns])


@responses.activate
def test_get_not_found(project_id, prediction_id, predictions_metadata_by_id_url):
    add_response(predictions_metadata_by_id_url, '', status=404)
    with pytest.raises(errors.ClientError):
        Predictions.get(project_id, prediction_id)


@responses.activate
def test_get_as_data_frame_server_error(project_id, prediction_id, predictions_metadata_by_id_url):
    add_response(predictions_metadata_by_id_url, '', status=500)
    with pytest.raises(errors.ServerError):
        Predictions.get(project_id, prediction_id)


def test_repr():
    preds = Predictions('project_id', 'prediction_id', 'model_id', 'dataset_id', False)

    assert repr(preds) == "Predictions(prediction_id='prediction_id', project_id='project_id', " \
                          "model_id='model_id', dataset_id='dataset_id', " \
                          "includes_prediction_intervals=False, prediction_intervals_size=None, " \
                          "forecast_point=None, predictions_start_date=None, " \
                          "predictions_end_date=None, actual_value_column=None, " \
                          "explanation_algorithm=None, max_explanations=None, shap_warnings=None)"



import json
from copy import deepcopy

import pytest
import responses

from datarobot import Deployment
from datarobot.enums import MODEL_REPLACEMENT_REASON
from datarobot.utils import from_api
from tests.utils import request_body_to_json


@responses.activate
@pytest.mark.usefixtures('deployment_get_response')
def test_model_replacement_validation(unittest_endpoint, deployment_data):
    deployment_id = deployment_data['id']
    status = 'passing'
    message = 'Model can be used to replace the current model of the deployment.'
    checks = {'targetName': {'status': 'passing', 'message': 'Target name matches.'}}

    url = '{}/deployments/{}/model/validation/'.format(unittest_endpoint, deployment_id)
    responses.add(
        responses.POST,
        url,
        body=json.dumps({'status': status, 'message': message, 'checks': checks}),
        status=200,
        content_type='application/json')

    deployment = Deployment.get(deployment_id)
    response = deployment.validate_replacement_model('5ca3936e0eb05d0155e2b7a5')

    request_body = request_body_to_json(responses.calls[1].request)
    assert request_body['modelId'] == '5ca3936e0eb05d0155e2b7a5'

    assert status == response[0]
    assert message == response[1]
    assert from_api(checks) == response[2]


@responses.activate
def test_model_replacement_submission_pred_int_disabled(unittest_endpoint, deployment_data):
    deployment_id = deployment_data['id']
    new_model_id = '5ca3936e0eb05d0155e2b7a5'
    model_replacement_url = '{}/deployments/{}/model/'.format(unittest_endpoint, deployment_id)
    deployment_url = '{}/deployments/{}/'.format(unittest_endpoint, deployment_id)
    status_url = '{}/status_url'.format(unittest_endpoint)
    settings_url = '{}/deployments/{}/settings/'.format(unittest_endpoint, deployment_id)

    new_deployment_data = deepcopy(deployment_data)
    new_deployment_data['model']['id'] = new_model_id
    responses.add(
        responses.GET,
        deployment_url,
        status=200,
        content_type='application/json',
        body=json.dumps(deployment_data))
    responses.add(
        responses.PATCH,
        model_replacement_url,
        headers={'Location': status_url},
        status=202)
    responses.add(
        responses.GET,
        status_url,
        headers={'Location': deployment_url},
        status=303)
    responses.add(
        responses.GET,
        deployment_url,
        status=200,
        content_type='application/json',
        body=json.dumps(new_deployment_data))
    responses.add(
        responses.GET,
        settings_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'predictionIntervals': {
            'enabled': False,
            'percentiles': []
        }})
    )

    deployment = Deployment.get(deployment_id)
    assert deployment.model['id'] == deployment_data['model']['id']
    deployment.replace_model(new_model_id, MODEL_REPLACEMENT_REASON.DATA_DRIFT)
    assert deployment.model['id'] == new_model_id


@responses.activate
def test_model_replacement_submission_pred_int_enabled(unittest_endpoint, deployment_data):
    deployment_id = deployment_data['id']
    new_model_id = '5ca3936e0eb05d0155e2b7a5'
    model_replacement_url = '{}/deployments/{}/model/'.format(unittest_endpoint, deployment_id)
    deployment_url = '{}/deployments/{}/'.format(unittest_endpoint, deployment_id)
    status_url = '{}/status_url'.format(unittest_endpoint)
    settings_url = '{}/deployments/{}/settings/'.format(unittest_endpoint, deployment_id)

    new_deployment_data = deepcopy(deployment_data)
    new_deployment_data['model']['id'] = new_model_id
    project_id = new_deployment_data['model']['projectId']
    responses.add(
        responses.GET,
        deployment_url,
        status=200,
        content_type='application/json',
        body=json.dumps(deployment_data))
    responses.add(
        responses.PATCH,
        model_replacement_url,
        headers={'Location': status_url},
        status=202)
    responses.add(
        responses.GET,
        status_url,
        headers={'Location': deployment_url},
        status=303)
    responses.add(
        responses.GET,
        deployment_url,
        status=200,
        content_type='application/json',
        body=json.dumps(new_deployment_data))
    responses.add(
        responses.GET,
        settings_url,
        status=200,
        content_type='application/json',
        body=json.dumps({'predictionIntervals': {
            'enabled': True,
            'percentiles': [80]
        }})
    )
    responses.add(
        responses.POST,
        '{}/projects/{}/models/{}/predictionIntervals/'.format(
            unittest_endpoint, project_id, new_model_id
        ),
        status=202,
        headers={'Location': '{}/projects/{}/jobs/1/'.format(unittest_endpoint, project_id)}
    )
    responses.add(
        responses.GET,
        '{}/projects/{}/jobs/1/'.format(unittest_endpoint, project_id),
        status=303,
        content_type='application/json',
        body=json.dumps({
            'status': u'COMPLETED', 'model_id': new_model_id, 'is_blocked': False,
            'url': '{}/projects/{}/jobs/1/'.format(unittest_endpoint, project_id),
            'job_type': 'calculate_prediction_intervals', 'project_id': project_id, 'id': 1
        }),
        headers={'Location': '{}/projects/{}/jobs/1/result/'.format(unittest_endpoint, project_id)}
    )
    responses.add(
        responses.PATCH,
        settings_url,
        headers={'Location': status_url},
        status=202
    )
    responses.add(
        responses.GET,
        status_url,
        headers={'Location': deployment_url},
        status=303
    )

    deployment = Deployment.get(deployment_id)
    assert deployment.model['id'] == deployment_data['model']['id']
    deployment.replace_model(new_model_id, MODEL_REPLACEMENT_REASON.DATA_DRIFT)
    assert deployment.model['id'] == new_model_id



def assert_version(version, version_json):
    assert version.id == version_json["id"]
    assert version.environment_id == version_json["environmentId"]
    assert version.label == version_json["label"]
    assert version.description == version_json["description"]
    assert version.build_status == version_json["buildStatus"]
    assert version.created_at == version_json["created"]


def assert_custom_model_version(version, version_json):
    assert version.id == version_json["id"]
    assert version.custom_model_id == version_json["customModelId"]
    assert version.label == version_json["label"]
    assert version.description == version_json["description"]
    assert version.version_minor == version_json["versionMinor"]
    assert version.version_major == version_json["versionMajor"]
    assert version.is_frozen == version_json["isFrozen"]

    assert len(version.items) == len(version_json["items"])
    for item, item_json in zip(version.items, version_json["items"]):
        assert item.id == item_json["id"]
        assert item.file_name == item_json["fileName"]
        assert item.file_path == item_json["filePath"]
        assert item.file_source == item_json["fileSource"]
        assert item.created_at == item_json["created"]

    assert version.created_at == version_json["created"]
